##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Bitwise expressions contain expressions for performing bitwise operations.
Most of these operations are equivalent to the
:mod:`Bitwise Operations API <aerospike_helpers.operations.bitwise_operations>`
for binary data.

Example::

    import aerospike_helpers.expressions as exp
    # Let blob bin "c" == bytearray([3] * 5).
    # Count set bits starting at 3rd byte in bin "c" to get count of 6.
    expr = exp.BitCount(16, 8 * 3, exp.BlobBin("c")).compile()
"""

# from __future__ import annotations
from typing import Union, Dict, Any
import aerospike
from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp
from aerospike_helpers.expressions.resources import _Keys
from aerospike_helpers.expressions.base import BlobBin

########################
# Bit Modify Expressions
########################

TypeBitValue = Union[bytes, bytearray]
TypeBinName = Union[_BaseExpr, str]
TypePolicy = Union[Dict[str, Any], None]


class BitResize(_BaseExpr):
    """Create an expression that performs a bit_resize operation."""

    _op = aerospike.OP_BIT_RESIZE

    def __init__(self, policy: "TypePolicy", byte_size: int, flags: int, bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            byte_size (int): Number of bytes the resulting blob should occupy.
            flags (int): One or a combination of bit resize flags.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Blob value expression of resized blob bin.

        Example::

            # Blob bin "c" == bytearray([1] * 5).
            # Resize blob bin "c" from the front so that the returned value is bytearray([0] * 5 + [1] * 5).
            expr = exp.BitResize(None, 10, aerospike.BIT_RESIZE_FROM_FRONT, exp.BlobBin("c")).compile()
        """
        self._children = (
            byte_size,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: flags} if flags is not None else {_Keys.VALUE_KEY: 0}
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitInsert(_BaseExpr):
    """Create an expression that performs a bit_insert operation."""

    _op = aerospike.OP_BIT_INSERT

    def __init__(self, policy: "TypePolicy", byte_offset: int, value: "TypeBitValue", bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            byte_offset (int): Integer byte index of where to insert the value.
            value (TypeBitValue): A bytes value or blob value expression to insert.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob containing the inserted bytes.

        Example::

            # Let blob bin "c" == bytearray([1] * 5).
            # Insert 3 so that returned value is bytearray([1, 3, 1, 1, 1, 1]).
            expr = exp.BitInsert(None, 1, bytearray([3]), exp.BlobBin("c")).compile()
        """
        self._children = (
            byte_offset,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitRemove(_BaseExpr):
    """Create an expression that performs a bit_remove operation."""

    _op = aerospike.OP_BIT_REMOVE

    def __init__(self, policy: "TypePolicy", byte_offset: int, byte_size: int, bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            byte_offset (int): Byte index of where to start removing from.
            byte_size (int): Number of bytes to remove.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob containing the remaining bytes.

        Example::

            # b = bytearray([1, 2, 3, 4, 5])
            expr = exp.BitRemove(None, 1, 1, exp.BlobBin("b")).compile()
            #   00000001 00000010 00000011 00000100 00000101
            #   0        1 (selected byte)
            #            xxxxxxxx (byte(s) to remove)
            #   ============================================
            #   00001001 <-- 00000001 00000001 00000001
        """
        self._children = (
            byte_offset,
            byte_size,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitSet(_BaseExpr):
    """Create an expression that performs a bit_set operation."""

    _op = aerospike.OP_BIT_SET

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, value: "TypeBitValue", bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start overwriting.
            bit_size (int): Number of bits to overwrite.
            value (TypeBitValue): Bytes value or blob expression containing bytes to write.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob expression with the bits overwritten.

        Example::

            # Let blob bin "c" == bytearray([0] * 5).
            # Set bit at offset 7 with size 1 bits to 1 to make the returned value bytearray([1, 0, 0, 0, 0]).
            expr = exp.BitSet(None, 7, 1, bytearray([255]), exp.BlobBin("c")).compile()
        """
        self._children = (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitOr(_BaseExpr):
    """Create an expression that performs a bit_or operation."""

    _op = aerospike.OP_BIT_OR

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, value: "TypeBitValue", bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            value (TypeBitValue): Bytes value or blob expression containing bytes to use in operation.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob with the bits operated on.

        Example::

            # Let blob bin "c" == bytearray([1] * 5).
            # bitwise Or `8` with the first byte of blob bin c so that the returned value is bytearray([9, 1, 1, 1, 1]).
            expr = exp.BitOr(None, 0, 8, bytearray([8]), exp.BlobBin("c")).compile()
            #   00000001 00000001 00000001 00000001 00000001
            #   0 (index)
            #   xxxxxxxx (bits applied)
            #   00001000 (OR operation)
            #   ============================================
            #   00001001 00000001 00000001 00000001 00000001

        """
        self._children = (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitXor(_BaseExpr):
    """Create an expression that performs a bit_xor operation."""

    _op = aerospike.OP_BIT_XOR

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, value: "TypeBitValue", bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            value (TypeBitValue): Bytes value or blob expression containing bytes to use in operation.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob with the bits operated on.

        Example::

            # Let blob bin "c" == bytearray([1] * 5).
            # bitwise Xor `1` with the first byte of blob bin c so that the returned value is bytearray([0, 1, 1, 1, 1])
            expr = exp.BitXor(None, 0, 8, bytearray([1]), exp.BlobBin("c")).compile()
        """
        self._children = (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitAnd(_BaseExpr):
    """Create an expression that performs a bit_and operation."""

    _op = aerospike.OP_BIT_AND

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, value: "TypeBitValue", bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            value (TypeBitValue): Bytes value or blob expression containing bytes to use in operation.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob with the bits operated on.

        Example::

            # Let blob bin "c" == bytearray([1] * 5).
            # bitwise and `0` with the first byte of blob bin c so that the returned value is bytearray([0, 1, 1, 1, 1])
            expr = exp.BitAnd(None, 0, 8, bytearray([0]), exp.BlobBin("c")).compile()
            #   00000001 00000001 00000001 00000001 00000001
            #   01234567 (offset)
            #   ^(+8)
            #   xxxxxxxx (bits selected)
            #   00000000 (& operation)
            #   ============================================
            #   00000000 00000001 00000001 00000001 00000001

        """
        self._children = (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitNot(_BaseExpr):
    """Create an expression that performs a bit_not operation."""

    _op = aerospike.OP_BIT_NOT

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob with the bits operated on.

        Example::

            # Let blob bin "c" == bytearray([255] * 5).
            # bitwise, not, all of "c" to get bytearray([254] * 5).
            expr = exp.BitNot(None, 0, 40, exp.BlobBin("c")).compile()
        """
        self._children = (
            bit_offset,
            bit_size,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitLeftShift(_BaseExpr):
    """Create an expression that performs a bit_lshift operation."""

    _op = aerospike.OP_BIT_LSHIFT

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, shift: int, bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            shift (int): Number of bits to shift by.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob with the bits operated on.

        Example::

            # Let blob bin "c" == bytearray([1] * 5).
            # Bit left shift the first byte of bin "c" to get bytearray([8, 1, 1, 1, 1]).
            expr = exp.BitLeftShift(None, 0, 8, 3, exp.BlobBin("c")).compile()
        """
        self._children = (
            bit_offset,
            bit_size,
            shift,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitRightShift(_BaseExpr):
    """Create an expression that performs a bit_rshift operation."""

    _op = aerospike.OP_BIT_RSHIFT

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, shift: int, bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            shift (int): Number of bits to shift by.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob with the bits operated on.

        Example::

            # Let blob bin "c" == bytearray([8] * 5).
            # Bit left shift the first byte of bin "c" to get bytearray([4, 8, 8, 8, 8]).
            expr = exp.BitRightShift(None, 0, 8, 1, exp.BlobBin("c")).compile()
            # 00001000 00001000 00001000 00001000 00001000
            # 0 (offset=0)
            # xxxxxxxx (selected bits -> by 1 only in that region)
            # The rest of the bits are unaffected
            # 00000100 00001000 00001000 00001000 00001000

        """
        self._children = (
            bit_offset,
            bit_size,
            shift,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitAdd(_BaseExpr):
    """Create an expression that performs a bit_add operation.
    Note: integers are stored big-endian.
    """

    _op = aerospike.OP_BIT_ADD

    def __init__(
        self, policy: "TypePolicy", bit_offset: int, bit_size: int, value: int, action: int, bin: "TypeBinName"
    ):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            value (int): Integer value or expression for value to add.
            action (int): An aerospike bit overflow action.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: resulting blob with the bits operated on.

        Example::

            # Assume we have a blob bin of five bytes: bytearray([1, 1, 1, 1, 1])
            expr = exp.BitAdd(None, 8, 8, 1, aerospike.BIT_OVERFLOW_FAIL, exp.BlobBin("b")).compile()
            # Treat the selected bits as a number and add the value to it
            #   00000001 00000001 00000001 00000001 00000001
            #   01234567 89012345 (offset)
            #            ^(+8)
            #            xxxxxxxx (bits selected)
            # + 00000000 00000001 (value to add)
            #   ============================================
            #   00000001 00000010 00000001 00000001 00000001
        """
        self._children = (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: action} if action is not None else {_Keys.VALUE_KEY: 0}
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitSubtract(_BaseExpr):
    """Create an expression that performs a bit_subtract operation.
    Note: integers are stored big-endian.
    """

    _op = aerospike.OP_BIT_SUBTRACT

    def __init__(
        self, policy: "TypePolicy", bit_offset: int, bit_size: int, value: int, action: int, bin: "TypeBinName"
    ):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start operation.
            bit_size (int): Number of bits to be operated on.
            value (int): Integer value or expression for value to add.
            action (int): An aerospike bit overflow action.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: resulting blob with the bits operated on.

        Example::

            # Let blob bin "c" == bytearray([1] * 5).
            # Bit subtract the second byte of bin "c" to get bytearray([1, 0, 1, 1, 1])
            expr = exp.BitSubtract(None, 8, 8, 1, aerospike.BIT_OVERFLOW_FAIL).compile()
        """
        self._children = (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS, 0, {_Keys.VALUE_KEY: action} if action is not None else {_Keys.VALUE_KEY: 0}
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


class BitSetInt(_BaseExpr):
    """Create an expression that performs a bit_set_int operation.
    Note: integers are stored big-endian.
    """

    _op = aerospike.OP_BIT_SET_INT

    def __init__(self, policy: "TypePolicy", bit_offset: int, bit_size: int, value: int, bin: "TypeBinName"):
        """Args:
            policy (TypePolicy): Optional dictionary of :ref:`Bit policies <aerospike_bit_policies>`.
            bit_offset (int): Bit index of where to start writing.
            bit_size (int): Number of bits to overwrite.
            value (int): Integer value or integer expression containing value to write.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Resulting blob expression with the bits overwritten.

        Example::

            # Let blob bin "c" == bytearray([0] * 5).
            # Set bit at offset 7 with size 1 bytes to 1 to make the returned value bytearray([1, 0, 0, 0, 0]).
            expr = exp.BitSetInt(None, 7, 1, 1, exp.BlobBin("c")).compile()
        """
        self._children = (
            bit_offset,
            bit_size,
            value,
            _GenericExpr(
                _ExprOp._AS_EXP_BIT_FLAGS,
                0,
                {_Keys.VALUE_KEY: policy["bit_write_flags"]}
                if policy is not None and "bit_write_flags" in policy
                else {_Keys.VALUE_KEY: 0},
            ),
            bin if isinstance(bin, _BaseExpr) else BlobBin(bin),
        )


######################
# Bit Read Expressions
######################


class BitGet(_BaseExpr):
    """Create an expression that performs a bit_get operation."""

    _op = aerospike.OP_BIT_GET

    def __init__(self, bit_offset: int, bit_size: int, bin: "TypeBinName"):
        """Args:
            bit_offset (int): Bit index of where to start reading.
            bit_size (int): Number of bits to get.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Blob, bit_size bits rounded up to the nearest byte size.

        Example::

            # Let blob bin "c" == bytearray([1, 2, 3, 4, 5).
            # Get 2 from bin "c".
            expr = exp.BitGet(8, 8, exp.BlobBin("c")).compile()
        """
        self._children = (bit_offset, bit_size, bin if isinstance(bin, _BaseExpr) else BlobBin(bin))


class BitCount(_BaseExpr):
    """Create an expression that performs a bit_count operation."""

    _op = aerospike.OP_BIT_COUNT

    def __init__(self, bit_offset: int, bit_size: int, bin: "TypeBinName"):
        """Args:
            bit_offset (int): Bit index of where to start reading.
            bit_size (int): Number of bits to count.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Blob, bit_size bits rounded up to the nearest byte size.

        Example::

            # Let blob bin "c" == bytearray([3] * 5).
            # Count set bits starting at 3rd byte in bin "c" to get count of 6.
            expr = exp.BitCount(16, 8 * 3, exp.BlobBin("c")).compile()
            # 00000011 00000011 00000011 00000011 00000011
            # 01234567 89012345 6 (offset = 16)
            #                   12345678 90123456 78901234 (bit count = 24)
            # Number of 1's = 6
        """
        self._children = (bit_offset, bit_size, bin if isinstance(bin, _BaseExpr) else BlobBin(bin))


class BitLeftScan(_BaseExpr):
    """Create an expression that performs a bit_lscan operation."""

    _op = aerospike.OP_BIT_LSCAN

    def __init__(self, bit_offset: int, bit_size: int, value: bool, bin: "TypeBinName"):
        """Args:
            bit_offset (int): Bit index of where to start reading.
            bit_size (int): Number of bits to read.
            value bool: Bit value to check for.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Index of the left most bit starting from bit_offset set to value. Returns -1 if not found.

        Example::

            # Let blob bin "c" == bytearray([3] * 5).
            # Scan the first byte of bin "c" for the first bit set to 1. (should get 6)
            expr = exp.BitLeftScan(0, 8, True, exp.BlobBin("c")).compile()
        """
        self._children = (bit_offset, bit_size, value, bin if isinstance(bin, _BaseExpr) else BlobBin(bin))


class BitRightScan(_BaseExpr):
    """Create an expression that performs a bit_rscan operation."""

    _op = aerospike.OP_BIT_RSCAN

    def __init__(self, bit_offset: int, bit_size: int, value: bool, bin: "TypeBinName"):
        """Args:
            bit_offset (int): Bit index of where to start reading.
            bit_size (int): Number of bits to read.
            value (bool): Bit value to check for.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Index of the right most bit starting from bit_offset set to value. Returns -1 if not found.

        Example::

            # b = bytearray([1, 0, 0, 0, 128])
            expr = exp.BitRightScan(32, 8, True, exp.BlobBin("b")).compile()
            # 00000001 00000000 00000000 00000000 10000000
            # 01234567 89012345 67890123 45678901 2 (offset=32)
            #                                     xxxxxxxx (selected bits)
            #                                     01234567 (local offset)
            # 1 found at local offset 0
        """
        self._children = (bit_offset, bit_size, value, bin if isinstance(bin, _BaseExpr) else BlobBin(bin))


class BitGetInt(_BaseExpr):
    """Create an expression that performs a bit_get_int operation."""

    _op = aerospike.OP_BIT_GET_INT

    def __init__(self, bit_offset: int, bit_size: int, sign: bool, bin: "TypeBinName"):
        """Args:
            bit_offset (int): Bit index of where to start reading.
            bit_size (int): Number of bits to get.
            sign bool: True for signed, False for unsigned.
            bin (TypeBinName): A :class:`~aerospike_helpers.expressions.base.BlobBin` expression.

        :return: Integer expression.

        Example::

            # Let blob bin "c" == bytearray([1, 2, 3, 4, 5).
            # Get 2 as an integer from bin "c".
            expr = exp.BitGetInt(8, 8, True, exp.BlobBin("c")).compile()
        """
        self._children = (bit_offset, bit_size, 1 if sign else 0, bin if isinstance(bin, _BaseExpr) else BlobBin(bin))



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Module with helper functions to create dictionaries used by:

* :mod:`aerospike.Client.operate` and :mod:`aerospike.Client.operate_ordered`
* Certain batch operations listed in :mod:`aerospike_helpers.batch.records`
"""
import warnings

import aerospike
from typing import Optional


def read(bin_name):
    """Create a read operation dictionary.

    The read operation reads and returns the value in `bin_name`.

    Args:
        bin_name (str): the name of the bin from which to read.
    Returns:
        A dictionary to be passed to operate or operate_ordered.
    """

    return {
        "op": aerospike.OPERATOR_READ,
        "bin": bin_name,
    }


def write(bin_name, write_item):
    """Create a write operation dictionary.

    The write operation writes `write_item` into the bin specified by bin_name.

    Args:
        bin_name (str): The name of the bin into which `write_item` will be stored.
        write_item: The value which will be written into the bin.
    Returns:
        A dictionary to be passed to operate or operate_ordered.
    """
    return {"op": aerospike.OPERATOR_WRITE, "bin": bin_name, "val": write_item}


def delete():
    """Create a delete operation dictionary.

    The delete operation deletes a record and all associated bins.
    Requires server version >= 4.7.0.8.

    Returns:
        A dictionary to be passed to operate or operate_ordered.
    """

    return {
        "op": aerospike.OPERATOR_DELETE,
    }


def append(bin_name, append_item):
    """Create an append operation dictionary.

    The append operation appends `append_item` to the value in bin_name.

    Args:
        bin_name (str): The name of the bin to be used.
        append_item: The value which will be appended to the item contained in the specified bin.
    Returns:
        A dictionary to be passed to operate or operate_ordered.
    """
    return {"op": aerospike.OPERATOR_APPEND, "bin": bin_name, "val": append_item}


def prepend(bin_name, prepend_item):
    """Create a prepend operation dictionary.

    The prepend operation prepends `prepend_item` to the value in bin_name.

    Args:
        bin_name (str): The name of the bin to be used.
        prepend_item: The value which will be prepended to the item contained in the specified bin.
    Returns:
        A dictionary to be passed to operate or operate_ordered.
    """
    return {"op": aerospike.OPERATOR_PREPEND, "bin": bin_name, "val": prepend_item}


def increment(bin_name, amount):
    """Create an increment operation dictionary.

    The increment operation increases a value in bin_name by the specified amount,
    or creates a bin with the value of amount.

    Args:
        bin_name (str): The name of the bin to be incremented.
        amount: The amount by which to increment the item in the specified bin.
    Returns:
        A dictionary to be passed to operate or operate_ordered.
    """
    return {"op": aerospike.OPERATOR_INCR, "bin": bin_name, "val": amount}


def touch(ttl: Optional[int] = None):
    """Create a touch operation dictionary.

    Using ttl here is deprecated. It should be set in the record metadata for the operate method.

    Args:
        ttl (int): Deprecated. The ttl that should be set for the record.
            This should be set in the metadata passed to the operate or
            operate_ordered methods.
    Returns:
        A dictionary to be passed to operate or operate_ordered.
    """
    op_dict = {"op": aerospike.OPERATOR_TOUCH}
    if ttl:
        warnings.warn("TTL should be specified in the meta dictionary for operate", DeprecationWarning)
        op_dict["val"] = ttl
    return op_dict



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
The expressions base module provide expressions for:

* declaring variables, using variables, and control-flow
* comparison operators
* applying logical operators to one or more 'boolean expressions'
* returning the value of (in-memory) record metadata
* returning the value from storage, such as bin data or the record's key

"""

# from __future__ import annotations
from typing import Union, Any
import aerospike
from aerospike_helpers.expressions.resources import _GenericExpr
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import _ExprOp
from aerospike_helpers.expressions.resources import ResultType
from aerospike_helpers.expressions.resources import _Keys

TypeComparisonArg = Union[_BaseExpr, Any]
TypeGeo = Union[_BaseExpr, aerospike.GeoJSON]


###################
# Value Expressions
###################

class Val(_BaseExpr):
    """
    Create an expression that returns a value.
    """

    _op = _ExprOp._AS_EXP_CODE_AS_VAL

    def __init__(self, value: Any):
        """:return: (value)
        """
        self._fixed = {_Keys.VALUE_KEY: value}
        super().__init__()


class Unknown(_BaseExpr):
    """Create an 'Unknown' expression, which allows an operation expression
    ('read expression' or 'write expression') to be aborted.

    This expression returns a special 'unknown' trilean value which, when
    returned by an operation expression, will result in an error code 26
    :class:`~aerospike.exception.OpNotApplicable`. These failures can be ignored with the policy flags
    :class:`aerospike.EXP_READ_EVAL_NO_FAIL` for read expressions and
    :class:`aerospike.EXP_WRITE_EVAL_NO_FAIL` for write expressions.
    This would then allow subsequent operations in the transaction to proceed.

    This expression is only useful from a
    :class:`~aerospike_helpers.expressions.base.Cond` conditional expression within
    an operation expression, and should be avoided in filter-expressions, where
    it might trigger an undesired move into the storage-data phase.

    If a test-expression within the ``Cond`` yields the special 'unknown' trilean
    value, then the ``Cond`` will also immediately yield the 'unknown' value and
    further test-expressions will not be evaluated.

    Note that this special 'unknown' trilean value is the same value returned
    by any failed expression.
    """

    _op = _ExprOp.UNKNOWN

    def __init__(self):
        """:return: (unknown value)

        Example::

            from aerospike_helpers.expressions.arithmetic import Add

            # Declare variable for balance bin
            exp.Let(exp.Def("bal", exp.IntBin("balance")),
                # If IntBin("balance") >= 50, get "balance" + 50.
                exp.Cond(
                    exp.GE(exp.Var("bal"), 50),
                        Add(exp.Var("bal"), 50),
                    # Otherwise, fail the expression via Unknown().
                    exp.Unknown()
                )
            )
        """
        super().__init__()


########################
# Record Key Expressions
########################


class _Key(_BaseExpr):
    _op = _ExprOp.REC_KEY


class KeyInt(_Key):
    """Create an expression that returns the key as an integer. Returns the unknown-value if
    the key is not an integer.
    """

    _rt = ResultType.INTEGER

    def __init__(self):
        """:return: (integer value): Integer value of the key if the key is an integer.

        Example::

            # Integer record key >= 10000.
            expr = exp.GE(exp.KeyInt(), 10000).compile()
        """
        super().__init__()


class KeyStr(_Key):
    """Create an expression that returns the key as a string. Returns the unknown-value if
    the key is not a string.
    """

    _rt = ResultType.STRING

    def __init__(self):
        """:return: (string value): string value of the key if the key is an string.

        Example::

            # string record key == "aaa".
            expr = exp.Eq(exp.KeyStr(), "aaa").compile()
        """
        super().__init__()


class KeyBlob(_Key):
    """Create an expression that returns the key as a blob. Returns the unknown-value if
    the key is not a blob.
    """

    _rt = ResultType.BLOB

    def __init__(self):
        """:return: (blob value): Blob value of the key if the key is a blob.

        Example::

            # blob record key <= bytearray([0x65, 0x65]).
            expr = exp.GE(exp.KeyBlob(), bytearray([0x65, 0x65])).compile()
        """
        super().__init__()


class KeyExists(_BaseExpr):
    """Create an expression that returns if the primary key is stored in the record storage
    data as a boolean expression. This would occur on record write, when write policies set the `key` field to
    :class:`aerospike.POLICY_KEY_SEND`.
    """

    _op = _ExprOp.META_KEY_EXISTS
    _rt = ResultType.BOOLEAN

    def __init__(self):
        """:return: (boolean value): True if the record has a stored key, false otherwise.

        Example::

            # Key exists in record meta data.
            expr = exp.KeyExists().compile()
        """
        super().__init__()


#################
# Bin Expressions
#################


class BoolBin(_BaseExpr):
    """Create an expression that returns a bin as a boolean. Returns the unknown-value
    if the bin is not a boolean.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.BOOLEAN

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (boolean bin)

        Example::

            # Boolean bin "a" is True.
            expr = exp.BoolBin("a").compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class IntBin(_BaseExpr):
    """Create an expression that returns a bin as an integer. Returns the unknown-value
    if the bin is not an integer.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.INTEGER

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (integer bin)

        Example::

            # Integer bin "a" == 200.
            expr = exp.Eq(exp.IntBin("a"), 200).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class StrBin(_BaseExpr):
    """Create an expression that returns a bin as a string. Returns the unknown-value
    if the bin is not a string.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.STRING

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (string bin)

        Example::

            # String bin "a" == "xyz".
            expr = exp.Eq(exp.StrBin("a"), "xyz").compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class FloatBin(_BaseExpr):
    """Create an expression that returns a bin as a float. Returns the unknown-value
    if the bin is not a float.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.FLOAT

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (float bin)

        Example::

            # Float bin "a" > 2.71.
            expr = exp.GT(exp.FloatBin("a"), 2.71).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class BlobBin(_BaseExpr):
    """Create an expression that returns a bin as a blob. Returns the unknown-value
    if the bin is not a blob.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.BLOB

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (blob bin)

        Example::

            #. Blob bin "a" == bytearray([0x65, 0x65])
            expr = exp.Eq(exp.BlobBin("a"), bytearray([0x65, 0x65])).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class GeoBin(_BaseExpr):
    """Create an expression that returns a bin as a geojson. Returns the unknown-value
    if the bin is not a geojson.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.GEOJSON

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (geojson bin)

        Example::

            #GeoJSON bin "a" contained by GeoJSON bin "b".
            expr = exp.CmpGeo(GeoBin("a"), exp.GeoBin("b")).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class ListBin(_BaseExpr):
    """Create an expression that returns a bin as a list. Returns the unknown-value
    if the bin is not a list.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.LIST

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (list bin)

        Example::

            from aerospike_helpers.expressions import list as list_exprs

            # Check that list bin "listBin" contains at least one item with value 42.
            list42Count = list_exprs.ListGetByValue(
                ctx=None,
                return_type=aerospike.LIST_RETURN_COUNT,
                value=42,
                bin=exp.ListBin("listBin")
            )
            expr = exp.GT(list42Count, 0).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class MapBin(_BaseExpr):
    """Create an expression that returns a bin as a map. Returns the unknown-value
    if the bin is not a map.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.MAP

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (map bin)

        Example::

            from aerospike_helpers.expressions import map as map_exprs

            # Map bin "a" size > 7.
            expr = exp.GT(map_exprs.MapSize(None, exp.MapBin("a")), 7).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class HLLBin(_BaseExpr):
    """Create an expression that returns a bin as a HyperLogLog. Returns the unknown-value
    if the bin is not a HyperLogLog.
    """

    _op = _ExprOp.BIN
    _rt = ResultType.HLL

    def __init__(self, bin: str):
        """Args:
            bin (str): Bin name.

        :return: (HyperLogLog bin)

        Example::

            # Does HLL bin "a" have a hll_count > 1000000.
            from aerospike_helpers.expressions import hll
            count = hll.HLLGetCount(hll.HLLBin("a"))
            expr = exp.GT(count, 1000000).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class BinExists(_BaseExpr):
    """Create an expression that returns True if bin exists."""

    _op = _ExprOp.BIN_EXISTS
    _rt = ResultType.BOOLEAN

    def __init__(self, bin: str):
        """Args:
            bin (str): bin name.

        :return: (boolean value): True if bin exists, False otherwise.

        Example::

            #Bin "a" exists in record.
            expr = exp.BinExists("a").compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


class BinType(_BaseExpr):
    """Create an expression that returns the type of a bin
    as one of the aerospike :ref:`bin types <aerospike_bin_types>`
    """

    _op = _ExprOp.BIN_TYPE
    _rt = ResultType.INTEGER

    def __init__(self, bin: str):
        """Args:
            bin (str): bin name.

        :return: (integer value): returns the bin type.

        Example::

            # bin "a" == type string.
            expr = exp.Eq(exp.BinType("a"), aerospike.AS_BYTES_STRING).compile()
        """
        self._fixed = {_Keys.BIN_KEY: bin}


####################
# Record Expressions
####################


class SetName(_BaseExpr):
    """Create an expression that returns record set name string.
    This expression usually evaluates quickly because record
    meta data is cached in memory.
    """

    _op = _ExprOp.META_SET_NAME
    _rt = ResultType.STRING

    def __init__(self):
        """:return: (string value): Name of the set this record belongs to.

        Example::

            # Record set name == "myset".
            expr = exp.Eq(exp.SetName(), "myset").compile()
        """
        super().__init__()


class DeviceSize(_BaseExpr):
    """Create an expression that returns record size on disk. If server storage-engine is
    memory, then zero is returned. This expression usually evaluates quickly
    because record meta data is cached in memory.

    This expression should only be used for server versions less than 7.0. Use
    :py:class:`RecordSize` for server version 7.0+.
    """

    _op = _ExprOp.META_DEVICE_SIZE
    _rt = ResultType.INTEGER

    def __init__(self):
        """:return: (integer value): Uncompressed storage size of the record.

        Example::

            # Record device size >= 100 KB.
            expr = exp.GE(exp.DeviceSize(), 100 * 1024).compile()
        """
        super().__init__()


class MemorySize(_BaseExpr):
    """Returns the records memory size in bytes as an integer when either the namespace
    is configured data-in-memory true or storage-engine memory, otherwise returns 0.

    Requires server version between 5.3 inclusive and 7.0 exclusive.
    Use :py:class:`RecordSize` for server version 7.0+.
    """

    _op = _ExprOp.META_MEMORY_SIZE
    _rt = ResultType.INTEGER

    def __init__(self):
        """:return: (integer value)
        """
        super().__init__()


class RecordSize(_BaseExpr):
    """
    Create expression that returns the record size. This expression usually evaluates
    quickly because record meta data is cached in memory.

    Requires server version 7.0+. This expression replaces :py:class:`DeviceSize()` and
    :py:class:`MemorySize()`.
    """
    _op = _ExprOp.META_RECORD_SIZE
    _rt = ResultType.INTEGER

    def __init__(self):
        """
        :return: (integer value) Record size in bytes
        """
        super().__init__()


class LastUpdateTime(_BaseExpr):
    """Create an expression that the returns record last update time expressed as 64 bit
    integer nanoseconds since 1970-01-01 epoch.
    """

    _op = _ExprOp.META_LAST_UPDATE_TIME
    _rt = ResultType.INTEGER

    def __init__(self):
        """:return: (integer value): When the record was last updated.

        Example::

            # Record last update time >= 2020-01-15.
            expr = exp.GE(exp.LastUpdateTime(), 1577836800).compile()
        """
        super().__init__()


class SinceUpdateTime(_BaseExpr):
    """Create an expression that returns milliseconds since the record was last updated.
    This expression usually evaluates quickly because record meta data is cached in memory.
    """

    _op = _ExprOp.META_SINCE_UPDATE_TIME
    _rt = ResultType.INTEGER

    def __init__(self):
        """:return: (integer value): Number of milliseconds since last updated.

        Example::

            # Record last updated more than 2 hours ago.
            expr = exp.GT(exp.SinceUpdateTime(), 2 * 60 * 60 * 1000).compile()
        """
        super().__init__()


class VoidTime(_BaseExpr):
    """Create an expression that returns record expiration time expressed as 64 bit
    integer nanoseconds since 1970-01-01 epoch.
    """

    _op = _ExprOp.META_VOID_TIME
    _rt = ResultType.INTEGER

    def __init__(self):
        """:return: (integer value): Expiration time in nanoseconds since 1970-01-01.

        Example::

            # Record expires on 2021-01-01.
            expr = exp.And(
                    exp.GE(exp.VoidTime(), 1609459200),
                    exp.LT(exp.VoidTime(), 1609545600)).compile()
        """
        super().__init__()


class TTL(_BaseExpr):
    """Create an expression that returns record expiration time (time to live) in integer
    seconds.
    """

    _op = _ExprOp.META_TTL
    _rt = ResultType.INTEGER

    def __init__(self):
        """:return: (integer value): Number of seconds till the record will expire,
                                returns -1 if the record never expires.

        Example::

            # Record expires in less than 1 hour.
            expr = exp.LT(exp.TTL(), 60 * 60).compile()
        """
        super().__init__()


class IsTombstone(_BaseExpr):
    """Create an expression that returns if record has been deleted and is still in
    tombstone state. This expression usually evaluates quickly because record
    meta data is cached in memory. NOTE: this is only applicable for XDR filter expressions.
    """

    _op = _ExprOp.META_IS_TOMBSTONE
    _rt = ResultType.BOOLEAN

    def __init__(self):
        """:return: (boolean value): True if the record is a tombstone, false otherwise.

        Example::

            # Detect deleted records that are in tombstone state.
            expr = exp.IsTombstone().compile()
        """
        super().__init__()


class DigestMod(_BaseExpr):
    """Create an expression that returns record digest modulo as integer."""

    _op = _ExprOp.META_DIGEST_MOD
    _rt = ResultType.INTEGER

    def __init__(self, mod: int):
        """Args:
            mod (int): Divisor used to divide the digest to get a remainder.

        :return: (integer value): Value in range 0 and mod (exclusive).

        Example::

            # Records that have digest(key) % 3 == 1.
            expr = exp.Eq(exp.DigestMod(3), 1).compile()
        """
        self._fixed = {_Keys.VALUE_KEY: mod}


########################
# Comparison Expressions
########################


class Eq(_BaseExpr):
    """Create an equals, (==) expression."""

    _op = _ExprOp.EQ

    def __init__(self, expr0: "TypeComparisonArg", expr1: "TypeComparisonArg"):
        """Args:
            expr0 (TypeComparisonArg): Left argument to `==`.
            expr1 (TypeComparisonArg): Right argument to `==`.

        :return: (boolean value)

        Example::

            # Integer bin "a" == 11
            expr = exp.Eq(exp.IntBin("a"), 11).compile()
        """
        self._children = (expr0, expr1)


class NE(_BaseExpr):
    """Create a not equals (not ==) expressions."""

    _op = _ExprOp.NE

    def __init__(self, expr0: "TypeComparisonArg", expr1: "TypeComparisonArg"):
        """Args:
            expr0 (TypeComparisonArg): Left argument to `not ==`.
            expr1 (TypeComparisonArg): Right argument to `not ==`.

        :return: (boolean value)

        Example::

            # Integer bin "a" not == 13.
            expr = exp.NE(exp.IntBin("a"), 13).compile()
        """
        self._children = (expr0, expr1)


class GT(_BaseExpr):
    """Create a greater than (>) expression."""

    _op = _ExprOp.GT

    def __init__(self, expr0: "TypeComparisonArg", expr1: "TypeComparisonArg"):
        """Args:
            expr0 (TypeComparisonArg): Left argument to `>`.
            expr1 (TypeComparisonArg): Right argument to `>`.

        :return: (boolean value)

        Example::

            # Integer bin "a" > 8.
            expr = exp.GT(exp.IntBin("a"), 8).compile()
        """
        self._children = (expr0, expr1)


class GE(_BaseExpr):
    """Create a greater than or equal to (>=) expression."""

    _op = _ExprOp.GE

    def __init__(self, expr0: "TypeComparisonArg", expr1: "TypeComparisonArg"):
        """Args:
            expr0 (TypeComparisonArg): Left argument to `>=`.
            expr1 (TypeComparisonArg): Right argument to `>=`.

        :return: (boolean value)

        Example::

            # Integer bin "a" >= 88.
            expr = exp.GE(exp.IntBin("a"), 88).compile()
        """
        self._children = (expr0, expr1)


class LT(_BaseExpr):
    """Create a less than (<) expression."""

    _op = _ExprOp.LT

    def __init__(self, expr0: "TypeComparisonArg", expr1: "TypeComparisonArg"):
        """Args:
            expr0 (TypeComparisonArg): Left argument to `<`.
            expr1 (TypeComparisonArg): Right argument to `<`.

        :return: (boolean value)

        Example::

            # Integer bin "a" < 1000.
            expr = exp.LT(exp.IntBin("a"), 1000).compile()
        """
        self._children = (expr0, expr1)


class LE(_BaseExpr):
    """Create a less than or equal to (<=) expression."""

    _op = _ExprOp.LE

    def __init__(self, expr0: "TypeComparisonArg", expr1: "TypeComparisonArg"):
        """Args:
            expr0 (TypeComparisonArg): Left argument to `<=`.
            expr1 (TypeComparisonArg): Right argument to `<=`.

        :return: (boolean value)

        Example::

            # Integer bin "a" <= 1.
            expr = exp.LE(exp.IntBin("a"), 1).compile()
        """
        self._children = (expr0, expr1)


class CmpRegex(_BaseExpr):
    """Create an expression that performs a regex match on a string bin or value expression."""

    _op = _ExprOp.CMP_REGEX

    def __init__(self, options: int, regex_str: str, cmp_str: Union[_BaseExpr, str]):
        """Args:
            options (int) :ref:`regex_constants`: One of the aerospike regex constants, :ref:`regex_constants`.
            regex_str (str): POSIX regex string.
            cmp_str (Union[_BaseExpr, str]): String expression to compare against.

        :return: (boolean value)

        Example::

            # Select string bin "a" that starts with "prefix" and ends with "suffix".
            # Ignore case and do not match newline.
            import aerospike
            expr = exp.CmpRegex(aerospike.REGEX_ICASE | aerospike.REGEX_NEWLINE, "prefix.*suffix", \
            exp.StrBin("a")).compile()
        """
        self._children = (cmp_str,)
        self._fixed = {_Keys.REGEX_OPTIONS_KEY: options, _Keys.VALUE_KEY: regex_str}


class CmpGeo(_BaseExpr):
    """Create a point within region or region contains point expression."""

    _op = _ExprOp.CMP_GEO

    def __init__(self, expr0: "TypeGeo", expr1: "TypeGeo"):
        """Args:
            expr0 (TypeGeo): Left expression in comparison.
            expr1 (TypeGeo): Right expression in comparison.

        :return: (boolean value)

        Example::

            # Geo bin "point" is within geo bin "region".
            expr = exp.CmpGeo(exp.GeoBin("point"), exp.GeoBin("region")).compile()
        """
        self._children = (expr0, expr1)


#####################
# Logical Expressions
#####################


class Not(_BaseExpr):
    """Create a "not" (not) operator expression."""

    _op = _ExprOp.NOT

    def __init__(self, *exprs: _BaseExpr):
        """Args:
            `*exprs` (_BaseExpr): Variable amount of expressions to be negated.

        :return: (boolean value)

        Example::

            # not (a == 0 or a == 10)
            expr = exp.Not(exp.Or(
                        exp.Eq(exp.IntBin("a"), 0),
                        exp.Eq(exp.IntBin("a"), 10))).compile()
        """
        self._children = exprs


class And(_BaseExpr):
    """Create an "and" operator that applies to a variable amount of expressions."""

    _op = _ExprOp.AND

    def __init__(self, *exprs: _BaseExpr):
        """Args:
            `*exprs` (_BaseExpr): Variable amount of expressions to be ANDed together.

        :return: (boolean value)

        Example::

            # (a > 5 || a == 0) && b < 3
            expr = exp.And(
                    exp.Or(
                      exp.GT(exp.IntBin("a"), 5),
                      exp.Eq(exp.IntBin("a"), 0)),
                    exp.LT(exp.IntBin("b"), 3)).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Or(_BaseExpr):
    """Create an "or" operator that applies to a variable amount of expressions."""

    _op = _ExprOp.OR

    def __init__(self, *exprs: _BaseExpr):
        """Args:
            `*exprs` (_BaseExpr): Variable amount of expressions to be ORed together.

        :return: (boolean value)

        Example::

            # (a == 0 || b == 0)
            expr = exp.Or(
                    exp.Eq(exp.IntBin("a"), 0),
                    exp.Eq(exp.IntBin("b"), 0)).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Exclusive(_BaseExpr):
    """Create an expression that returns True if only one of the expressions are True."""

    _op = _ExprOp.EXCLUSIVE

    def __init__(self, *exprs: _BaseExpr):
        """Args:
            `*exprs` (_BaseExpr): Variable amount of expressions to be checked.

        :return: (boolean value)

        Example::

            # exclusive(a == 0, b == 0)
            expr = exp.Exclusive(
                            exp.Eq(exp.IntBin("a"), 0),
                            exp.Eq(exp.IntBin("b"), 0)).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


#######################################
# Flow Control and Variable Expressions
#######################################


class Cond(_BaseExpr):
    """Conditionally select an expression from a variable number of
    condition/action pairs, followed by a default expression action.

    Takes a set of test-expression/action-expression pairs and evaluates
    each test expression, one at a time. If a test returns ``True``, ``Cond``
    evaluates the corresponding action expression and returns its value,
    after which ``Cond`` doesn't evaluate any of the other tests or
    expressions.  If all tests evaluate to ``False``, the default action
    expression is evaluated and returned.

    ``Cond`` is strictly typed, so all actions-expressions must evaluate to
    the same type or the :class:`~aerospike_helpers.expressions.base.Unknown` expression.

    Requires server version 5.6.0+.
    """

    _op = _ExprOp.COND

    def __init__(self, *exprs: _BaseExpr):
        """Args:
            `*exprs` (_BaseExpr): bool exp1, action exp1, bool exp2, action exp2, ..., action-default

        :return: (boolean value)

        Example::

            from aerospike_helpers.expressions.arithmetic import Add, Sub, Mul

            import aerospike
            # Configure the client
            config = {
                'hosts': [('127.0.0.1', 3000)]
            }

            # Create a client and connect it to the cluster
            client = aerospike.client(config)
            client.truncate('test', "demo", 0)

            # Store 2 bin integers and use expressions to perform arithmetic
            # Results will only be calculated and returned, not stored

            keyTuple = ('test', 'demo', 'key')
            client.put(keyTuple, {"operation": "add", "val1": 40, "val2": 30})

            # Determine operation to perform
            # If operation is unknown, return -1
            expr = exp.Cond(
                        exp.Eq(exp.StrBin("operation"), "add"),
                            Add(exp.IntBin("val1"), exp.IntBin("val2")),
                        exp.Eq(exp.StrBin("operation"), "subtract"),
                            Sub(exp.IntBin("val1"), exp.IntBin("val2")),
                        exp.Eq(exp.StrBin("operation"), "multiply"),
                            Mul(exp.IntBin("val1"), exp.IntBin("val2")),
            -1).compile()

            from aerospike_helpers.operations import expression_operations as expr_ops
            ops = [
                # Bin "results" doesn't actually exist in the server
                # The name is only used to return the results
                expr_ops.expression_read("results", expr)
            ]
            record = client.operate(keyTuple, ops)
            print(record)
            # (('test', 'demo', 'key', bytearray(b'...')), {'ttl': 2592000, 'gen': 1}, {'results': 70})

            client.put(keyTuple, {"operation": "divide"})

            record = client.operate(keyTuple, ops)
            print(record)
            # Divide isn't supported, so we get -1
            # (('test', 'demo', 'key', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'results': -1})
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Let(_BaseExpr):
    """Defines variables to be used within the ``Let`` expression's scope. The last
    argument can be any expression and should make use of the defined
    variables. The ``Let`` expression returns the evaluated result of the last
    argument. This expression is useful if you need to reuse the result of a
    complicated or expensive expression.
    """

    _op = _ExprOp.LET

    def __init__(self, *exprs: _BaseExpr):
        """Args:
            `*exprs` (_BaseExpr): Variable number of :class:`~aerospike_helpers.expressions.base.Def` expressions
            followed by a scoped expression.

        :return: (result of scoped expression)

        Example::

            # for int bin "a", 5 < a < 10
            expr = exp.Let(exp.Def("x", exp.IntBin("a")),
                exp.And(
                    exp.GT(5, exp.Var("x")),
                    exp.LT(exp.Var("x"), 10))).compile()
        """
        self._children = exprs + (_GenericExpr(_ExprOp._AS_EXP_CODE_END_OF_VA_ARGS, 0, {}),)


class Def(_BaseExpr):
    """Assign variable to an expression that can be accessed later with
    :class:`~aerospike_helpers.expressions.base.Var`.
    Requires server version 5.6.0+.
    """

    _op = _ExprOp.DEF

    def __init__(self, var_name: str, expr: _BaseExpr):
        """Args:
            `var_name` (str): Variable name.
            `expr` (_BaseExpr): Variable is set to result of this expression.

        :return: (a variable name expression pair)

        Example::

            # for int bin "a", 5 < a < 10
            expr = exp.Let(exp.Def("x", exp.IntBin("a")),
                    exp.And(
                        exp.LT(5, exp.Var("x")),
                        exp.LT(exp.Var("x"), 10))).compile()
        """
        self._fixed = {_Keys.VALUE_KEY: var_name}
        self._children = (expr,)


class Var(_BaseExpr):
    """Retrieve expression value from a variable previously defined with
    :class:`~aerospike_helpers.expressions.base.Def`.
    Requires server version 5.6.0+.
    """

    _op = _ExprOp.VAR

    def __init__(self, var_name: str):
        """Args:
            `var_name` (str): Variable name.

        :return: (value stored in variable)

        Example::

            # for int bin "a", 5 < a < 10
            expr = exp.Let(exp.Def("x", exp.IntBin("a")),
                    exp.And(
                        exp.LT(5, exp.Var("x")),
                        exp.LT(exp.Var("x"), 10))).compile()
        """
        self._fixed = {_Keys.VALUE_KEY: var_name}



##########################################################################
# Copyright 2013-2022 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

import typing as ty
from typing import Optional

TypeOps = ty.List[ty.Dict]
TypeBatchPolicyWrite = ty.Union[ty.Dict, None]
TypeBatchPolicyRemove = ty.Union[ty.Dict, None]
TypeBatchPolicyApply = ty.Union[ty.Dict, None]
TypeBatchPolicyRead = ty.Union[ty.Dict, None]
TypeRecord = ty.Union[ty.Tuple, None]
TypeUDFArgs = ty.List[ty.Any]


class _Types:
    READ = 0
    WRITE = 1
    APPLY = 2
    REMOVE = 3


class BatchRecord:
    """ BatchRecord provides the base fields for BatchRecord objects.

        BatchRecord should usually be read from as a result and not created by the user. Its subclasses can be used as
        input to batch_write.
        Client methods :meth:`~Client.batch_apply`, :meth:`~Client.batch_operate`, :meth:`~Client.batch_remove`
        with batch_records field as a list of these BatchRecord objects containing the batch request results.

        Attributes:
            key (:obj:`tuple`): The aerospike key to operate on.
            record (:ref:`aerospike_record_tuple`): The record corresponding to the requested key.
            result (int): The status code of the operation.
            in_doubt (bool): Is it possible that the write transaction completed even though an error was generated. \
            This may be the case when a client error occurs (like timeout) after the command was sent \
            to the server.
    """

    def __init__(self, key: tuple) -> None:
        self.key = key
        self.record = None
        self.result = 0
        self.in_doubt = False


class Write(BatchRecord):
    """ Write is used for executing Batch write operations with batch_write and retrieving batch write results.

        Attributes:
            key (:obj:`tuple`): The aerospike key to operate on.
            record (:obj:`tuple`): The record corresponding to the requested key.
            result (int): The status code of the operation.
            in_doubt (bool): Is it possible that the write transaction completed even though an error was generated. \
            This may be the case when a client error occurs (like timeout) after the command was sent \
            to the server.
            ops (:ref:`aerospike_operation_helpers.operations`): A list of aerospike operation dictionaries to perform
                on the record at key.
            meta (dict): the metadata to set for the operations in this BatchRecord
            policy (:ref:`aerospike_batch_write_policies`, optional): An optional dictionary of batch write policy
                flags.
    """

    def __init__(
        self, key: tuple, ops: "TypeOps", meta: Optional[dict] = None, policy: "TypeBatchPolicyWrite" = None
    ) -> None:
        """
        Example::

            # Create a batch Write to increment bin "a" by 10 and read the result from the record.
            import aerospike
            import aerospike_helpers.operations as op
            from aerospike_helpers.batch.records import Write

            bin_name = "a"

            namespace = "test"
            set = "demo"
            user_key = 1
            key = (namespace, set, user_key)

            ops = [
                op.increment(bin_name, 10),
                op.read(bin_name)
            ]

            meta={"gen": 1, "ttl": aerospike.TTL_NEVER_EXPIRE}
            bw = Write(key, ops, meta=meta)
        """
        super().__init__(key)
        self.ops = ops
        self._type = _Types.WRITE
        self._has_write = True
        self.meta = meta
        self.policy = policy


class Read(BatchRecord):
    """ Read is used for executing Batch read operations with batch_write and retrieving results.

        Attributes:
            key (:obj:`tuple`): The aerospike key to operate on.
            record (:obj:`tuple`): The record corresponding to the requested key.
            result (int): The status code of the operation.
            in_doubt (bool): Is it possible that the write transaction completed even though an error was generated. \
            This may be the case when a client error occurs (like timeout) after the command was sent \
            to the server.
            ops (:ref:`aerospike_operation_helpers.operations`): list of aerospike operation dictionaries to perform on
                the record at key.
            meta (dict): the metadata to set for the operations in this BatchRecord
            read_all_bins (bool, optional): An optional bool, if True, read all bins in the record.
            policy (:ref:`aerospike_batch_read_policies`, optional): An optional dictionary of batch read policy flags.
    """

    def __init__(
        self,
        key: tuple,
        ops: ty.Union[TypeOps, None],
        read_all_bins: bool = False,
        meta: Optional[dict] = None,
        policy: "TypeBatchPolicyRead" = None,
    ) -> None:
        """
        Example::

            # Create a batch Read to read bin "a" from the record.
            import aerospike
            import aerospike_helpers.operations as op
            from aerospike_helpers.batch.records import Read

            bin_name = "a"

            namespace = "test"
            set = "demo"
            user_key = 1
            key = (namespace, set, user_key)

            ops = [
                op.read(bin_name)
            ]

            meta={"gen": 1, "ttl": aerospike.TTL_NEVER_EXPIRE}
            br = Read(key, ops, meta=meta)
        """
        super().__init__(key)
        self.ops = ops
        self.read_all_bins = read_all_bins
        self._type = _Types.READ
        self._has_write = False
        self.meta = meta
        self.policy = policy


class Apply(BatchRecord):
    """ BatchApply is used for executing Batch UDF (user defined function) apply operations with batch_write and
        retrieving results.

        Attributes:
            key (:obj:`tuple`): The aerospike key to operate on.
            module (str): Name of the lua module previously registered with the server.
            function (str): Name of the UDF to invoke.
            args (:obj:`list`): List of arguments to pass to the UDF.
            record (:ref:`aerospike_record_tuple`): The record corresponding to the requested key.
            result (int): The status code of the operation.
            in_doubt (bool): Is it possible that the write transaction completed even though an error was generated. \
            This may be the case when a client error occurs (like timeout) after the command was sent \
            to the server.
            policy (:ref:`aerospike_batch_apply_policies`, optional): An optional dictionary of batch apply policy
                flags.
    """

    def __init__(
        self, key: tuple, module: str, function: str, args: "TypeUDFArgs", policy: "TypeBatchPolicyApply" = None
    ) -> None:
        """
        Example::

            # Create a batch Apply to apply UDF "test_func" to bin "a" from the record.
            # Assume that "test_func" takes a bin name string as an argument.
            # Assume the appropriate UDF module has already been registered.
            import aerospike_helpers.operations as op


            module = "my_lua"
            function = "test_func"

            bin_name = "a"
            args = [
                bin_name
            ]

            namespace = "test"
            set = "demo"
            user_key = 1
            key = (namespace, set, user_key)

            ba = Apply(key, module, function, args)
        """
        super().__init__(key)
        self._type = _Types.APPLY
        self._has_write = True
        self.module = module
        self.function = function
        self.args = args
        self.policy = policy


class Remove(BatchRecord):
    """ Remove is used for executing Batch remove operations with batch_write and retrieving results.

        Attributes:
            key (:obj:`tuple`): The aerospike key to operate on.
            record (:ref:`aerospike_record_tuple`): The record corresponding to the requested key.
            result (int): The status code of the operation.
            in_doubt (bool): Is it possible that the write transaction completed even though an error was generated. \
            This may be the case when a client error occurs (like timeout) after the command was sent \
            to the server.
            policy (:ref:`aerospike_batch_remove_policies`, optional): An optional dictionary of batch remove policy
                flags.
    """

    def __init__(self, key: tuple, policy: "TypeBatchPolicyRemove" = None) -> None:
        """
        Example::

            # Create a batch Remove to remove the record.
            import aerospike_helpers.operations as op


            namespace = "test"
            set = "demo"
            user_key = 1
            key = (namespace, set, user_key)

            br = Remove(key, ops)
        """
        super().__init__(key)
        self._type = _Types.REMOVE
        self._has_write = True
        self.policy = policy


TypeBatchRecordList = ty.List[BatchRecord]


class BatchRecords:
    """ BatchRecords is used as input and output for multiple batch APIs.

        Attributes:
            batch_records (list): A list of BatchRecord subtype objects used to \
            define batch operations and hold results. BatchRecord Types can be Remove, Write, \
            Read, and Apply.
            result (int): The status code of the last batch call that used this BatchRecords.
                ``0`` if all batch subtransactions succeeded (or if the only failures were \
                    ``FILTERED_OUT`` or ``RECORD_NOT_FOUND``)
                Not ``0`` if an error occurred. The most common error is ``-16`` \
                    (One or more batch sub transactions failed).
    """

    def __init__(self, batch_records: Optional[TypeBatchRecordList] = None) -> None:
        """
        Example::

            import aerospike
            import aerospike_helpers.operations.operations as op
            from aerospike_helpers.batch.records import BatchRecords, Remove, Write, Read

            # Setup
            config = {
                "hosts": [("127.0.0.1", 3000)]
            }
            client = aerospike.client(config)

            namespace = "test"
            set_ = "demo"
            keys = [
                (namespace, set_, 1),
                (namespace, set_, 2),
                (namespace, set_, 3),
            ]
            bin_name = "id"
            for key in keys:
                client.put(key, {bin_name: 1})

            # Create a BatchRecords to remove a record, write a bin, and read a bin.
            brs = BatchRecords(
                [
                    Remove(
                        key=keys[0],
                    ),
                    Write(
                        key=keys[1],
                        ops=[
                            op.write(bin_name, 100),
                            op.read(bin_name),
                        ]
                    ),
                    Read(
                        key=keys[2],
                        ops=[
                            op.read(bin_name)
                        ]
                    )
                ]
            )

            # Note this call will mutate brs and set results in it.
            client.batch_write(brs)
            for br in brs.batch_records:
                print(br.result)
                print(br.record)
            # 0
            # (('test', 'demo', 1, bytearray(b'...')), {'ttl': 4294967295, 'gen': 0}, {})
            # 0
            # (('test', 'demo', 2, bytearray(b'...')), {'ttl': 2592000, 'gen': 4}, {'id': 100})
            # 0
            # (('test', 'demo', 3, bytearray(b'...')), {'ttl': 2592000, 'gen': 3}, {'id': 1})
        """

        if batch_records is None:
            batch_records = []

        self.batch_records = batch_records
        self.result = 0



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
"""
Helper functions to create HyperLogLog operation dictionary arguments for:

* :mod:`aerospike.Client.operate` and :mod:`aerospike.Client.operate_ordered`
* Certain batch operations listed in :mod:`aerospike_helpers.batch.records`

HyperLogLog bins and operations allow for your application to form fast, reasonable approximations
of members in the union or intersection between multiple HyperLogLog bins.
HyperLogLog’s estimates are a balance between complete accuracy and efficient savings
in space and speed in dealing with extremely large datasets.

    .. note:: HyperLogLog operations require server version >= 4.9.0

    .. seealso:: `HyperLogLog (Data Type) more info. \
        <https://docs.aerospike.com/server/guide/data-types/hll#operations>`_.

Example::

    import aerospike
    from aerospike_helpers.operations import hll_operations as hll_ops
    from aerospike_helpers.operations import operations

    NUM_INDEX_BITS = 12
    NUM_MH_BITS = 24

    # Configure the client.
    config = {"hosts": [("127.0.0.1", 3000)]}
    # Create a client and connect it to the cluster.
    client = aerospike.client(config)

    # Create customer keys
    TEST_NS = "test"
    TEST_SET = "demo"
    customerNames = ["Amy", "Farnsworth", "Scruffy"]
    keys = []
    for customer in customerNames:
        keys.append((TEST_NS, TEST_SET, customer))

    itemsViewedPerCustomer = [
        # [item1, item2, ... item500]
        list("item%s" % str(i) for i in range(0, 500)), # Amy
        list("item%s" % str(i) for i in range(0, 750)), # Farnsworth
        list("item%s" % str(i) for i in range(250, 1000)), # Scruffy
    ]

    for key, itemsViewed in zip(keys, itemsViewedPerCustomer):
        customerName = key[2]
        ops = [
            operations.write("name", customerName),
            hll_ops.hll_add("viewed", itemsViewed, NUM_INDEX_BITS, NUM_MH_BITS),
        ]
        client.operate(key, ops)

    # Find out how many items viewed Amy, Farnsworth, and Scruffy have in common.
    farnsworthRecord = client.get(keys[1])
    scruffyRecord = client.get(keys[2])
    farnsworthViewedItems = farnsworthRecord[2]["viewed"]
    scruffyViewedItems = scruffyRecord[2]["viewed"]
    viewed = [farnsworthViewedItems, scruffyViewedItems]
    ops = [
        hll_ops.hll_get_intersect_count("viewed", viewed)
    ]
    # Pass in Amy's key
    _, _, res = client.operate(keys[0], ops)
    print("Estimated items viewed intersection:", res["viewed"])
    # Estimated items viewed intersection: 251
    # Actual intersection: 250

    # Find out how many unique products Amy, Farnsworth, and Scruffy have viewed.
    ops = [hll_ops.hll_get_union_count("viewed", viewed)]
    _, _, res = client.operate(keys[0], ops)

    print("Estimated items viewed union:", res["viewed"])
    # Estimated items viewed union: 1010
    # Actual union: 1000

    # Find the similarity of Amy, Farnsworth, and Scruffy's product views.
    ops = [hll_ops.hll_get_similarity("viewed", viewed)]
    _, _, res = client.operate(keys[0], ops)

    print("Estimated items viewed similarity: %f%%" % (res["viewed"] * 100))
    # Estimated items viewed similarity: 24.888393%
    # Actual similarity: 25%

"""

import aerospike


OP_KEY = "op"
BIN_KEY = "bin"
HLL_POLICY_KEY = "hll_policy"
INDEX_BIT_COUNT_KEY = "index_bit_count"
MH_BIT_COUNT_KEY = "mh_bit_count"
VALUE_LIST_KEY = "value_list"


def hll_add(bin_name: str, values, index_bit_count=None, mh_bit_count=None, policy=None):
    """Creates a hll_add operation.

    Server will add the values to the hll bin.
    If the HLL bin does not exist, it will be created with index_bit_count and/or mh_bit_count if they have been
    supplied.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        values: The values to be added to the HLL set.
        index_bit_count: An optional number of index bits. Must be between 4 and 16 inclusive.
        mh_bit_count: An optional number of min hash bits. Must be between 4 and 58 inclusive.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_ADD,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: values,
        INDEX_BIT_COUNT_KEY: -1 if index_bit_count is None else index_bit_count,
        MH_BIT_COUNT_KEY: -1 if mh_bit_count is None else mh_bit_count,
    }

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict


def hll_describe(bin_name):
    """Creates a hll_describe operation.

    Server returns index and minhash bit counts used to create HLL bin in a list of integers.
    The list size is 2.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_DESCRIBE,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_fold(bin_name: str, index_bit_count):
    """Creates a hll_fold operation.

    Servers folds index_bit_count to the specified value.
    This can only be applied when minhash bit count on the HLL bin is 0.
    Server does not return a value.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index_bit_count: number of index bits. Must be between 4 and 16 inclusive.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_FOLD, BIN_KEY: bin_name, INDEX_BIT_COUNT_KEY: index_bit_count}

    return op_dict


def hll_get_count(bin_name):
    """Creates a hll_get_count operation.

    Server returns estimated count of elements in the HLL bin.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_GET_COUNT,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_get_intersect_count(bin_name: str, hll_list):
    """Creates a hll_get_intersect_count operation.

    Server returns estimate of elements that would be contained by the intersection of these HLL objects.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be intersected.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_INTERSECT_COUNT, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_get_similarity(bin_name: str, hll_list):
    """Creates a hll_get_similarity operation.

    Server returns estimated similarity of the HLL objects.
    Server returns a float.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs used for similarity estimation.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_SIMILARITY, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_get_union(bin_name: str, hll_list):
    """Creates a hll_get_union operation.

    Server returns an HLL object that is the union of all specified HLL objects
    in hll_list with the HLL bin.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be unioned.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_UNION, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_get_union_count(bin_name: str, hll_list):
    """Creates a hll_get_union_count operation.

    Server returns the estimated count of elements that would be contained by the union of all specified HLL objects
    in the list with the HLL bin.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be unioned.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_GET_UNION_COUNT, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    return op_dict


def hll_init(bin_name: str, index_bit_count=None, mh_bit_count=None, policy=None):
    """Creates a hll_init operation.

    Server creates a new HLL or resets an existing HLL.
    If index_bit_count and mh_bit_count are None, an existing HLL bin will be reset but retain its configuration.
    If 1 of index_bit_count or mh_bit_count are set,
    an existing HLL bin will set that config and retain its current value for the unset config.
    If the HLL bin does not exist, index_bit_count is required to create it, mh_bit_count is optional.
    Server does not return a value.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index_bit_count: An optional number of index bits. Must be between 4 and 16 inclusive.
        mh_bit_count: An optional number of min hash bits. Must be between 4 and 58 inclusive.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_INIT,
        BIN_KEY: bin_name,
        INDEX_BIT_COUNT_KEY: -1 if index_bit_count is None else index_bit_count,
        MH_BIT_COUNT_KEY: -1 if mh_bit_count is None else mh_bit_count,
    }

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict


def hll_refresh_count(bin_name: str):
    """Creates a hll_refresh_count operation.

    Server updates the cached count if it is stale.
    Server returns the count.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.
    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_REFRESH_COUNT,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_set_union(bin_name: str, hll_list, policy=None):
    """Creates a hll_set_union operation.

    Server sets the union of all specified HLL objects with the HLL bin.
    Server returns nothing.

    Returns a dictionary to be used with :meth:`aerospike.Client.operate` and :meth:`aerospike.Client.operate_ordered`.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs who's union will be set.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {OP_KEY: aerospike.OP_HLL_SET_UNION, BIN_KEY: bin_name, VALUE_LIST_KEY: hll_list}

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict



from typing import Any, Callable, Union, final, Literal, Optional

from aerospike_helpers.batch.records import BatchRecords
from aerospike_helpers.metrics import MetricsPolicy

AS_BOOL: Literal[1]
AS_BYTES_BLOB: Literal[4]
AS_BYTES_BOOL: Literal[17]
AS_BYTES_CSHARP: Literal[8]
AS_BYTES_DOUBLE: Literal[2]
AS_BYTES_ERLANG: Literal[12]
AS_BYTES_GEOJSON: Literal[23]
AS_BYTES_HLL: Literal[18]
AS_BYTES_INTEGER: Literal[1]
AS_BYTES_JAVA: Literal[7]
AS_BYTES_LIST: Literal[20]
AS_BYTES_MAP: Literal[19]
AS_BYTES_PHP: Literal[11]
AS_BYTES_PYTHON: Literal[9]
AS_BYTES_RUBY: Literal[10]
AS_BYTES_STRING: Literal[3]
AS_BYTES_TYPE_MAX: Literal[24]
AS_BYTES_UNDEF: Literal[0]
AUTH_EXTERNAL: Literal[1]
AUTH_EXTERNAL_INSECURE: Literal[2]
AUTH_INTERNAL: Literal[0]
AUTH_PKI: Literal[3]
BIT_OVERFLOW_FAIL: Literal[0]
BIT_OVERFLOW_SATURATE: Literal[2]
BIT_OVERFLOW_WRAP: Literal[4]
BIT_RESIZE_DEFAULT: Literal[0]
BIT_RESIZE_FROM_FRONT: Literal[1]
BIT_RESIZE_GROW_ONLY: Literal[2]
BIT_RESIZE_SHRINK_ONLY: Literal[4]
BIT_WRITE_CREATE_ONLY: Literal[1]
BIT_WRITE_DEFAULT: Literal[0]
BIT_WRITE_NO_FAIL: Literal[4]
BIT_WRITE_PARTIAL: Literal[8]
BIT_WRITE_UPDATE_ONLY: Literal[2]
CDT_CTX_LIST_INDEX: Literal[0x10]
CDT_CTX_LIST_INDEX_CREATE: Literal[0x14]
CDT_CTX_LIST_RANK: Literal[0x11]
CDT_CTX_LIST_VALUE: Literal[0x13]
CDT_CTX_MAP_INDEX: Literal[0x20]
CDT_CTX_MAP_KEY: Literal[0x22]
CDT_CTX_MAP_KEY_CREATE: Literal[0x24]
CDT_CTX_MAP_RANK: Literal[0x21]
CDT_CTX_MAP_VALUE: Literal[0x23]
EXP_READ_DEFAULT: Literal[0]
EXP_READ_EVAL_NO_FAIL: Literal[16]
EXP_WRITE_ALLOW_DELETE: Literal[4]
EXP_WRITE_CREATE_ONLY: Literal[1]
EXP_WRITE_DEFAULT: Literal[0]
EXP_WRITE_EVAL_NO_FAIL: Literal[16]
EXP_WRITE_POLICY_NO_FAIL: Literal[8]
EXP_WRITE_UPDATE_ONLY: Literal[2]
HLL_WRITE_ALLOW_FOLD: Literal[8]
HLL_WRITE_CREATE_ONLY: Literal[1]
HLL_WRITE_DEFAULT: Literal[0]
HLL_WRITE_NO_FAIL: Literal[4]
HLL_WRITE_UPDATE_ONLY: Literal[2]
INDEX_BLOB: Literal[3]
INDEX_GEO2DSPHERE: Literal[2]
INDEX_NUMERIC: Literal[1]
INDEX_STRING: Literal[0]
INDEX_TYPE_DEFAULT: Literal[0]
INDEX_TYPE_LIST: Literal[1]
INDEX_TYPE_MAPKEYS: Literal[2]
INDEX_TYPE_MAPVALUES: Literal[3]
INTEGER: Literal[0]
JOB_QUERY: Literal["query"]
JOB_SCAN: Literal["scan"]
JOB_STATUS_COMPLETED: Literal[2]
JOB_STATUS_INPROGRESS: Literal[1]
JOB_STATUS_UNDEF: Literal[0]
LIST_ORDERED: Literal[1]
LIST_RETURN_COUNT: Literal[5]
LIST_RETURN_EXISTS: Literal[13]
LIST_RETURN_INDEX: Literal[1]
LIST_RETURN_NONE: Literal[0]
LIST_RETURN_RANK: Literal[3]
LIST_RETURN_REVERSE_INDEX: Literal[2]
LIST_RETURN_REVERSE_RANK: Literal[4]
LIST_RETURN_VALUE: Literal[7]
LIST_SORT_DEFAULT: Literal[0]
LIST_SORT_DROP_DUPLICATES: Literal[2]
LIST_UNORDERED: Literal[0]
LIST_WRITE_DEFAULT: Literal[0]
LIST_WRITE_ADD_UNIQUE: Literal[1]
LIST_WRITE_INSERT_BOUNDED: Literal[2]
LIST_WRITE_NO_FAIL: Literal[4]
LIST_WRITE_PARTIAL: Literal[8]
LOG_LEVEL_DEBUG: Literal[3]
LOG_LEVEL_ERROR: Literal[0]
LOG_LEVEL_INFO: Literal[2]
LOG_LEVEL_OFF: Literal[-1]
LOG_LEVEL_TRACE: Literal[4]
LOG_LEVEL_WARN: Literal[1]
MAP_KEY_ORDERED: Literal[1]
MAP_KEY_VALUE_ORDERED: Literal[3]
MAP_RETURN_COUNT: Literal[5]
MAP_RETURN_EXISTS: Literal[13]
MAP_RETURN_INDEX: Literal[1]
MAP_RETURN_KEY: Literal[6]
MAP_RETURN_KEY_VALUE: Literal[8]
MAP_RETURN_NONE: Literal[0]
MAP_RETURN_RANK: Literal[3]
MAP_RETURN_REVERSE_INDEX: Literal[2]
MAP_RETURN_REVERSE_RANK: Literal[4]
MAP_RETURN_VALUE: Literal[7]
MAP_RETURN_ORDERED_MAP: Literal[17]
MAP_RETURN_UNORDERED_MAP: Literal[16]
MAP_UNORDERED: Literal[0]
MAP_WRITE_FLAGS_CREATE_ONLY: Literal[1]
MAP_WRITE_FLAGS_DEFAULT: Literal[0]
MAP_WRITE_FLAGS_NO_FAIL: Literal[4]
MAP_WRITE_FLAGS_PARTIAL: Literal[8]
MAP_WRITE_FLAGS_UPDATE_ONLY: Literal[2]
MAP_WRITE_NO_FAIL: Literal[4]
MAP_WRITE_PARTIAL: Literal[8]
OPERATOR_APPEND: Literal[9]
OPERATOR_DELETE: Literal[14]
OPERATOR_INCR: Literal[6]
OPERATOR_PREPEND: Literal[10]
OPERATOR_READ: Literal[0]
OPERATOR_TOUCH: Literal[11]
OPERATOR_WRITE: Literal[1]
OP_BIT_ADD: Literal[2010]
OP_BIT_AND: Literal[2006]
OP_BIT_COUNT: Literal[2015]
OP_BIT_GET: Literal[2014]
OP_BIT_GET_INT: Literal[2012]
OP_BIT_INSERT: Literal[2001]
OP_BIT_LSCAN: Literal[2016]
OP_BIT_LSHIFT: Literal[2008]
OP_BIT_NOT: Literal[2007]
OP_BIT_OR: Literal[2004]
OP_BIT_REMOVE: Literal[2002]
OP_BIT_RESIZE: Literal[2000]
OP_BIT_RSCAN: Literal[2017]
OP_BIT_RSHIFT: Literal[2009]
OP_BIT_SET: Literal[2003]
OP_BIT_SET_INT: Literal[2013]
OP_BIT_SUBTRACT: Literal[2011]
OP_BIT_XOR: Literal[2005]
OP_EXPR_READ: Literal[2200]
OP_EXPR_WRITE: Literal[2201]
OP_HLL_ADD: Literal[2100]
OP_HLL_DESCRIBE: Literal[2101]
OP_HLL_FOLD: Literal[2102]
OP_HLL_GET_COUNT: Literal[2103]
OP_HLL_GET_INTERSECT_COUNT: Literal[2104]
OP_HLL_GET_SIMILARITY: Literal[2105]
OP_HLL_GET_UNION: Literal[2106]
OP_HLL_GET_UNION_COUNT: Literal[2107]
OP_HLL_INIT: Literal[2108]
OP_HLL_MAY_CONTAIN: Literal[2111]
OP_HLL_REFRESH_COUNT: Literal[2109]
OP_HLL_SET_UNION: Literal[2110]
OP_LIST_APPEND: Literal[1001]
OP_LIST_APPEND_ITEMS: Literal[1002]
OP_LIST_CLEAR: Literal[1009]
OP_LIST_GET: Literal[1011]
OP_LIST_GET_BY_INDEX: Literal[1016]
OP_LIST_GET_BY_INDEX_RANGE: Literal[1017]
OP_LIST_GET_BY_INDEX_RANGE_TO_END: Literal[1035]
OP_LIST_GET_BY_RANK: Literal[1018]
OP_LIST_GET_BY_RANK_RANGE: Literal[1019]
OP_LIST_GET_BY_RANK_RANGE_TO_END: Literal[1036]
OP_LIST_GET_BY_VALUE: Literal[1020]
OP_LIST_GET_BY_VALUE_LIST: Literal[1021]
OP_LIST_GET_BY_VALUE_RANGE: Literal[1022]
OP_LIST_GET_BY_VALUE_RANK_RANGE_REL: Literal[1033]
OP_LIST_GET_BY_VALUE_RANK_RANGE_REL_TO_END: Literal[1034]
OP_LIST_GET_RANGE: Literal[1012]
OP_LIST_INCREMENT: Literal[1015]
OP_LIST_INSERT: Literal[1003]
OP_LIST_INSERT_ITEMS: Literal[1004]
OP_LIST_POP: Literal[1005]
OP_LIST_POP_RANGE: Literal[1006]
OP_LIST_REMOVE: Literal[1007]
OP_LIST_REMOVE_BY_INDEX: Literal[1023]
OP_LIST_REMOVE_BY_INDEX_RANGE: Literal[1024]
OP_LIST_REMOVE_BY_INDEX_RANGE_TO_END: Literal[1039]
OP_LIST_REMOVE_BY_RANK: Literal[1025]
OP_LIST_REMOVE_BY_RANK_RANGE: Literal[1026]
OP_LIST_REMOVE_BY_RANK_RANGE_TO_END: Literal[1040]
OP_LIST_REMOVE_BY_REL_RANK_RANGE: Literal[1038]
OP_LIST_REMOVE_BY_REL_RANK_RANGE_TO_END: Literal[1037]
OP_LIST_REMOVE_BY_VALUE: Literal[1027]
OP_LIST_REMOVE_BY_VALUE_LIST: Literal[1028]
OP_LIST_REMOVE_BY_VALUE_RANGE: Literal[1029]
OP_LIST_REMOVE_BY_VALUE_RANK_RANGE_REL: Literal[1032]
OP_LIST_REMOVE_RANGE: Literal[1008]
OP_LIST_SET: Literal[1010]
OP_LIST_SET_ORDER: Literal[1030]
OP_LIST_SIZE: Literal[1014]
OP_LIST_SORT: Literal[1031]
OP_LIST_TRIM: Literal[1013]
OP_LIST_CREATE: Literal[1041]
OP_MAP_CLEAR: Literal[1107]
OP_MAP_DECREMENT: Literal[1105]
OP_MAP_GET_BY_INDEX: Literal[1122]
OP_MAP_GET_BY_INDEX_RANGE: Literal[1123]
OP_MAP_GET_BY_INDEX_RANGE_TO_END: Literal[1142]
OP_MAP_GET_BY_KEY: Literal[1118]
OP_MAP_GET_BY_KEY_INDEX_RANGE_REL: Literal[1131]
OP_MAP_GET_BY_KEY_LIST: Literal[1127]
OP_MAP_GET_BY_KEY_RANGE: Literal[1119]
OP_MAP_GET_BY_KEY_REL_INDEX_RANGE: Literal[1140]
OP_MAP_GET_BY_KEY_REL_INDEX_RANGE_TO_END: Literal[1136]
OP_MAP_GET_BY_RANK: Literal[1124]
OP_MAP_GET_BY_RANK_RANGE: Literal[1125]
OP_MAP_GET_BY_RANK_RANGE_TO_END: Literal[1143]
OP_MAP_GET_BY_VALUE: Literal[1120]
OP_MAP_GET_BY_VALUE_LIST: Literal[1126]
OP_MAP_GET_BY_VALUE_RANGE: Literal[1121]
OP_MAP_GET_BY_VALUE_RANK_RANGE_REL: Literal[1130]
OP_MAP_GET_BY_VALUE_RANK_RANGE_REL_TO_END: Literal[1141]
OP_MAP_INCREMENT: Literal[1104]
OP_MAP_PUT: Literal[1102]
OP_MAP_PUT_ITEMS: Literal[1103]
OP_MAP_REMOVE_BY_INDEX: Literal[1114]
OP_MAP_REMOVE_BY_INDEX_RANGE: Literal[1115]
OP_MAP_REMOVE_BY_INDEX_RANGE_TO_END: Literal[1134]
OP_MAP_REMOVE_BY_KEY: Literal[1108]
OP_MAP_REMOVE_BY_KEY_INDEX_RANGE_REL: Literal[1129]
OP_MAP_REMOVE_BY_KEY_LIST: Literal[1109]
OP_MAP_REMOVE_BY_KEY_RANGE: Literal[1110]
OP_MAP_REMOVE_BY_KEY_REL_INDEX_RANGE: Literal[1137]
OP_MAP_REMOVE_BY_KEY_REL_INDEX_RANGE_TO_END: Literal[1132]
OP_MAP_REMOVE_BY_RANK: Literal[1116]
OP_MAP_REMOVE_BY_RANK_RANGE: Literal[1117]
OP_MAP_REMOVE_BY_RANK_RANGE_TO_END: Literal[1135]
OP_MAP_REMOVE_BY_VALUE: Literal[1111]
OP_MAP_REMOVE_BY_VALUE_LIST: Literal[1112]
OP_MAP_REMOVE_BY_VALUE_RANGE: Literal[1113]
OP_MAP_REMOVE_BY_VALUE_RANK_RANGE_REL: Literal[1128]
OP_MAP_REMOVE_BY_VALUE_REL_INDEX_RANGE: Literal[1138]
OP_MAP_REMOVE_BY_VALUE_REL_RANK_RANGE: Literal[1139]
OP_MAP_REMOVE_BY_VALUE_REL_RANK_RANGE_TO_END: Literal[1133]
OP_MAP_CREATE: Literal[1144]
OP_MAP_SET_POLICY: Literal[1101]
OP_MAP_SIZE: Literal[1106]
POLICY_COMMIT_LEVEL_ALL: Literal[0]
POLICY_COMMIT_LEVEL_MASTER: Literal[1]
POLICY_EXISTS_CREATE: Literal[1]
POLICY_EXISTS_CREATE_OR_REPLACE: Literal[4]
POLICY_EXISTS_IGNORE: Literal[0]
POLICY_EXISTS_REPLACE: Literal[3]
POLICY_EXISTS_UPDATE: Literal[2]
POLICY_GEN_EQ: Literal[1]
POLICY_GEN_GT: Literal[2]
POLICY_GEN_IGNORE: Literal[0]
POLICY_KEY_DIGEST: Literal[0]
POLICY_KEY_SEND: Literal[1]
POLICY_READ_MODE_AP_ALL: Literal[1]
POLICY_READ_MODE_AP_ONE: Literal[0]
POLICY_READ_MODE_SC_ALLOW_REPLICA: Literal[2]
POLICY_READ_MODE_SC_ALLOW_UNAVAILABLE: Literal[3]
POLICY_READ_MODE_SC_LINEARIZE: Literal[1]
POLICY_READ_MODE_SC_SESSION: Literal[0]
POLICY_REPLICA_ANY: Literal[1]
POLICY_REPLICA_MASTER: Literal[0]
POLICY_REPLICA_PREFER_RACK: Literal[3]
POLICY_REPLICA_SEQUENCE: Literal[2]
POLICY_RETRY_NONE: Literal[0]
POLICY_RETRY_ONCE: Literal[1]
PRIV_DATA_ADMIN: Literal[2]
PRIV_READ: Literal[10]
PRIV_READ_WRITE: Literal[11]
PRIV_READ_WRITE_UDF: Literal[12]
PRIV_SINDEX_ADMIN: Literal[4]
PRIV_SYS_ADMIN: Literal[1]
PRIV_TRUNCATE: Literal[14]
PRIV_UDF_ADMIN: Literal[3]
PRIV_USER_ADMIN: Literal[0]
PRIV_WRITE: Literal[13]
REGEX_EXTENDED: Literal[1]
REGEX_ICASE: Literal[2]
REGEX_NEWLINE: Literal[8]
REGEX_NONE: Literal[0]
REGEX_NOSUB: Literal[4]
SERIALIZER_JSON: Literal[2]
SERIALIZER_NONE: Literal[0]
SERIALIZER_USER: Literal[3]
TTL_DONT_UPDATE: Literal[0xFFFFFFFE]
TTL_NAMESPACE_DEFAULT: Literal[0]
TTL_NEVER_EXPIRE: Literal[0xFFFFFFFF]
TTL_CLIENT_DEFAULT: Literal[0xFFFFFFFD]
UDF_TYPE_LUA: Literal[0]
QUERY_DURATION_LONG: Literal[0]
QUERY_DURATION_SHORT: Literal[1]
QUERY_DURATION_LONG_RELAX_AP: Literal[2]

@final
class CDTInfinite:
    def __init__(self, *args, **kwargs) -> None: ...

@final
class CDTWildcard:
    def __init__(self, *args, **kwargs) -> None: ...

class Client:
    def __init__(self, *args, **kwargs) -> None: ...
    def admin_change_password(self, username: str, password: str, policy: dict = ...) -> None: ...
    def admin_create_role(self, role: str, privileges: list, policy: dict = ..., whitelist: list = ..., read_quota: int = ..., write_quota: int = ...) -> None: ...
    def admin_create_user(self, username: str, password: str, roles: list, policy: dict = ...) -> None: ...
    def admin_drop_role(self, role: str, policy: dict = ...) -> None: ...
    def admin_drop_user(self, username: str, policy: dict = ...) -> None: ...
    def admin_get_role(self, role: str, policy: dict = ...) -> dict: ...
    def admin_get_roles(self, policy: dict = ...) -> dict: ...
    def admin_grant_privileges(self, role: str, privileges: list, policy: dict = ...) -> None: ...
    def admin_grant_roles(self, username: str, roles: list, policy: dict = ...) -> None: ...
    def admin_query_role(self, role: str, policy: dict = ...) -> list: ...
    def admin_query_roles(self, policy: dict = ...) -> dict: ...
    def admin_query_user(self, username: str, policy: dict = ...) -> list: ...
    def admin_query_user_info(self, *args, **kwargs) -> Any: ...
    def admin_query_users(self, policy: dict = ...) -> dict: ...
    def admin_query_users_info(self, *args, **kwargs) -> Any: ...
    def admin_revoke_privileges(self, role: str, privileges: list, policy: dict = ...) -> None: ...
    def admin_revoke_roles(self, username: str, roles: list, policy: dict = ...) -> None: ...
    def admin_set_password(self, username: str, password: str, policy: dict = ...) -> None: ...
    def admin_set_quotas(self, role: str, read_quota: int = ..., write_quota: int = ..., policy: dict = ...) -> None: ...
    def admin_set_whitelist(self, role: str, whitelist: list, policy: dict = ...) -> None: ...
    def append(self, key: tuple, bin: str, val: str, meta: dict = ..., policy: dict = ...) -> None: ...
    def apply(self, key: tuple, module: str, function: str, args: list, policy: dict = ...) -> Union[str, int, float, bytearray, list, dict]: ...
    def batch_apply(self, keys: list, module: str, function: str, args: list, policy_batch: dict = ..., policy_batch_apply: dict = ...) -> BatchRecords: ...
    def batch_get_ops(self, keys: list, ops: list, policy: dict) -> list: ...
    def batch_operate(self, keys: list, ops: list, policy_batch: dict = ..., policy_batch_write: dict = ..., ttl: int = ...) -> BatchRecords: ...
    def batch_remove(self, keys: list, policy_batch: dict = ..., policy_batch_remove: dict = ...) -> BatchRecords: ...
    def batch_read(self, keys: list, bins: list[str] = ..., policy_batch: dict = ...) -> BatchRecords: ...
    def batch_write(self, batch_records: BatchRecords, policy_batch: dict = ...) -> BatchRecords: ...
    def close(self) -> None: ...
    def connect(self, username: str = ..., password: str = ...) -> Client: ...
    def exists(self, key: tuple, policy: dict = ...) -> tuple: ...
    def exists_many(self, keys: list, policy: dict = ...) -> list: ...
    def get(self, key: tuple, policy: dict = ...) -> tuple: ...
    def get_cdtctx_base64(self, ctx: list) -> str: ...
    def get_expression_base64(self, expression) -> str: ...
    def get_key_partition_id(self, ns, set, key) -> int: ...
    def get_many(self, keys: list, policy: dict = ...) -> list: ...
    def get_node_names(self) -> list: ...
    def get_nodes(self) -> list: ...
    def increment(self, key: tuple, bin: str, offset: int, meta: dict = ..., policy: dict = ...) -> None: ...
    def index_cdt_create(self, *args, **kwargs) -> Any: ...
    def index_geo2dsphere_create(self, ns: str, set: str, bin: str, name: str, policy: dict = ...) -> None: ...
    def index_integer_create(self, ns: str, set: str, bin: str, name: str, policy: dict = ...) -> None: ...
    def index_list_create(self, ns: str, set: str, bin: str, index_datatype, name: str, policy: dict = ...) -> None: ...
    def index_map_keys_create(self, ns: str, set: str, bin: str, index_datatype, name: str, policy: dict = ...) -> None: ...
    def index_map_values_create(self, ns: str, set: str, bin: str, index_datatype, name: str, policy: dict = ...) -> None: ...
    def index_remove(self, ns, name: str, policy: dict = ...) -> None: ...
    def index_string_create(self, ns: str, set: str, bin: str, name: str, policy: dict = ...) -> None: ...
    def index_blob_create(self, ns: str, set: str, bin: str, name: str, policy: dict = ...) -> None: ...
    def info_all(self, command: str, policy: dict = ...) -> dict: ...
    def info_random_node(self, command: str, policy: dict = ...) -> str: ...
    def info_single_node(self, command: str, host: str, policy: dict = ...) -> str: ...
    def is_connected(self) -> bool: ...
    def job_info(self, job_id: int, module, policy: dict = ...) -> dict: ...
    def enable_metrics(self, policy: Optional[MetricsPolicy] = None) -> None: ...
    def disable_metrics(self) -> None: ...
    # List and map operations in the aerospike module are deprecated and undocumented
    # def list_append(self, *args, **kwargs) -> Any: ...
    # def list_clear(self, *args, **kwargs) -> Any: ...
    # def list_extend(self, *args, **kwargs) -> Any: ...
    # def list_get(self, *args, **kwargs) -> Any: ...
    # def list_get_range(self, *args, **kwargs) -> Any: ...
    # def list_insert(self, *args, **kwargs) -> Any: ...
    # def list_insert_items(self, *args, **kwargs) -> Any: ...
    # def list_pop(self, *args, **kwargs) -> Any: ...
    # def list_pop_range(self, *args, **kwargs) -> Any: ...
    # def list_remove(self, *args, **kwargs) -> Any: ...
    # def list_remove_range(self, *args, **kwargs) -> Any: ...
    # def list_set(self, *args, **kwargs) -> Any: ...
    # def list_size(self, *args, **kwargs) -> Any: ...
    # def list_trim(self, *args, **kwargs) -> Any: ...
    # def map_clear(self, *args, **kwargs) -> Any: ...
    # def map_decrement(self, *args, **kwargs) -> Any: ...
    # def map_get_by_index(self, *args, **kwargs) -> Any: ...
    # def map_get_by_index_range(self, *args, **kwargs) -> Any: ...
    # def map_get_by_key(self, *args, **kwargs) -> Any: ...
    # def map_get_by_key_list(self, *args, **kwargs) -> Any: ...
    # def map_get_by_key_range(self, *args, **kwargs) -> Any: ...
    # def map_get_by_rank(self, *args, **kwargs) -> Any: ...
    # def map_get_by_rank_range(self, *args, **kwargs) -> Any: ...
    # def map_get_by_value(self, *args, **kwargs) -> Any: ...
    # def map_get_by_value_list(self, *args, **kwargs) -> Any: ...
    # def map_get_by_value_range(self, *args, **kwargs) -> Any: ...
    # def map_increment(self, *args, **kwargs) -> Any: ...
    # def map_put(self, *args, **kwargs) -> Any: ...
    # def map_put_items(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_index(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_index_range(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_key(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_key_list(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_key_range(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_rank(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_rank_range(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_value(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_value_list(self, *args, **kwargs) -> Any: ...
    # def map_remove_by_value_range(self, *args, **kwargs) -> Any: ...
    # def map_set_policy(self, key, bin, map_policy) -> Any: ...
    # def map_size(self, *args, **kwargs) -> Any: ...
    def operate(self, key: tuple, list: list, meta: dict = ..., policy: dict = ...) -> tuple: ...
    def operate_ordered(self, key: tuple, list: list, meta: dict = ..., policy: dict = ...) -> list: ...
    def prepend(self, key: tuple, bin: str, val: str, meta: dict = ..., policy: dict = ...) -> None: ...
    def put(self, key: tuple, bins: dict, meta: dict = ..., policy: dict = ..., serializer = ...) -> None: ...
    def query(self, namespace: str, set: str = ...) -> Query: ...
    def query_apply(self, ns: str, set: str, predicate: tuple, module: str, function: str, args: list = ..., policy: dict = ...) -> int: ...
    def remove(self, key: tuple, meta: dict = ..., policy: dict = ...) -> None: ...
    def remove_bin(self, key: tuple, list: list, meta: dict = ..., policy: dict = ...) -> None: ...
    def scan(self, namespace: str, set: str = ...) -> Scan: ...
    def scan_apply(self, ns: str, set: str, module: str, function: str, args: list = ..., policy: dict = ..., options: dict = ...) -> int: ...
    def select(self, *args, **kwargs) -> tuple: ...
    def select_many(self, keys: list, bins: list, policy: dict = ...) -> list: ...
    def set_xdr_filter(self, data_center: str, namespace: str, expression_filter, policy: dict = ...) -> str: ...
    def shm_key(self) -> Union[int, None]: ...
    def touch(self, key: tuple, val: int = ..., meta: dict = ..., policy: dict = ...) -> None: ...
    def truncate(self, namespace: str, set: str, nanos: int, policy: dict = ...) -> int: ...
    def udf_get(self, module: str, language: int = ..., policy: dict = ...) -> str: ...
    def udf_list(self, policy: dict = ...) -> list: ...
    def udf_put(self, filename: str, udf_type = ..., policy: dict = ...) -> None: ...
    def udf_remove(self, module: str, policy: dict = ...) -> None: ...

class GeoJSON:
    geo_data: Any
    def __init__(self, geo_data: Union[str, dict] = ...) -> None: ...
    def dumps(self) -> str: ...
    def loads(self, raw_geo: str) -> None: ...
    def unwrap(self) -> dict: ...
    def wrap(self, geo_data: dict) -> None: ...

class KeyOrderedDict(dict):
    def __init__(self, *args, **kwargs) -> None: ...

class Query:
    max_records: int
    records_per_second: int
    ttl: int
    def __init__(self, *args, **kwargs) -> None: ...
    def add_ops(self, ops: list) -> None: ...
    def apply(self, module: str, function: str, arguments: list = ...) -> Any: ...
    def execute_background(self, policy: dict = ...) -> int: ...
    def foreach(self, callback: Callable, policy: dict = ..., options: dict = ...) -> None: ...
    def get_partitions_status(self) -> tuple: ...
    def is_done(self) -> bool: ...
    def paginate(self) -> None: ...
    def results(self, policy: dict = ..., options: dict = ...) -> list: ...
    # TODO: this isn't an infinite list of bins
    def select(self, *args, **kwargs) -> None: ...
    def where(self, predicate: tuple, ctx: list = ...) -> None: ...

class Scan:
    ttl: int
    def __init__(self, *args, **kwargs) -> None: ...
    def add_ops(self, ops: list) -> None: ...
    def apply(self, module: str, function: str, arguments: list = ...) -> Any: ...
    def foreach(self, callback: Callable, policy: dict = ..., options: dict = ..., nodename: str = ...) -> None: ...
    def execute_background(self, policy: dict = ...) -> int: ...
    def get_partitions_status(self) -> tuple: ...
    def is_done(self) -> bool: ...
    def paginate(self) -> None: ...
    def results(self, policy: dict = ..., nodename: str = ...) -> list: ...
    # TODO: this isn't an infinite list of bins
    def select(self, *args, **kwargs) -> None: ...

@final
class null:
    def __init__(self, *args, **kwargs) -> None: ...

def calc_digest(ns: str, set: str, key: Union[str, int, bytearray]) -> bytearray: ...
def client(config: dict) -> Client: ...
def geodata(geo_data: dict) -> GeoJSON: ...
def geojson(geojson_str: str) -> GeoJSON: ...
def get_cdtctx_base64(ctx: list) -> str: ...
def get_expression_base64(expression) -> str: ...
def get_partition_id(*args, **kwargs) -> Any: ...
def set_deserializer(callback: Callable) -> None: ...
def set_log_handler(callback: Callable = ...) -> None: ...
def set_log_level(log_level: int) -> None: ...
def set_serializer(callback: Callable) -> None: ...
def unset_serializers() -> None: ...



# -*- coding: utf-8 -*-
################################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from __future__ import print_function
import os
import platform
import sys
from subprocess import Popen
from subprocess import call, run
from setuptools import setup, Extension
from distutils.command.build import build
from distutils.command.clean import clean
from multiprocessing import cpu_count
import time
import io
import xml.etree.ElementTree as ET
import glob

################################################################################
# ENVIRONMENT VARIABLES
################################################################################
machine = platform.machine()
os.putenv('ARCHFLAGS', '-arch ' + machine)
os.environ['ARCHFLAGS'] = '-arch ' + machine
AEROSPIKE_C_VERSION = os.getenv('AEROSPIKE_C_VERSION')
BASEPATH = os.path.dirname(os.path.abspath(__file__))
AEROSPIKE_C_HOME = os.path.join(BASEPATH, 'aerospike-client-c')

AEROSPIKE_C_TARGET = None
PLATFORM = platform.platform(1)
LINUX = 'Linux' in PLATFORM
DARWIN = 'Darwin' in PLATFORM or 'macOS' in PLATFORM
WINDOWS = 'Windows' in PLATFORM

CWD = os.path.abspath(os.path.dirname(__file__))
STATIC_SSL = os.getenv('STATIC_SSL')
SSL_LIB_PATH = os.getenv('SSL_LIB_PATH')
# COVERAGE environment variable only meant for CI/CD workflow to generate C coverage data
# Not for developers to use, unless you know what the workflow is doing!
COVERAGE = os.getenv('COVERAGE')

# Applies no optimizations on both the C client and Python client
UNOPTIMIZED = os.getenv('UNOPTIMIZED')

# Include debug information on macOS (not included by default)
INCLUDE_DSYM = os.getenv('INCLUDE_DSYM')

################################################################################
# GENERIC BUILD SETTINGS
################################################################################

include_dirs = ['src/include'] + \
    [x for x in os.getenv('CPATH', '').split(':') if len(x) > 0] + \
    ['/usr/local/opt/openssl/include'] + \
    ['aerospike-client-c/modules/common/src/include']
extra_compile_args = [
    '-std=gnu99', '-g', '-Wall', '-fPIC', '-DDEBUG', '-O1',
    '-fno-common', '-fno-strict-aliasing',
    '-D_FILE_OFFSET_BITS=64', '-D_REENTRANT',
    '-DMARCH_' + machine,
]

if not WINDOWS:
    # Windows does not have this flag
    extra_compile_args.append("-Wno-strict-prototypes")
    extra_compile_args.append('-Wno-implicit-function-declaration')

if machine == 'x86_64':
    extra_compile_args.append('-march=nocona')
extra_objects = []

extra_link_args = []

SANITIZER=os.getenv('SANITIZER')
if SANITIZER:
    sanitizer_flags = [
        '-fsanitize=address',
        '-fsanitize-recover=all'
    ]
    extra_compile_args.extend(sanitizer_flags)

    extra_link_args.append("-static-libasan")
    extra_link_args.extend(sanitizer_flags)

library_dirs = ['/usr/local/opt/openssl/lib', '/usr/local/lib']
libraries = [
    'ssl',
    'crypto',
    'pthread',
    'm',
    'z'
]

##########################
# GITHUB ACTIONS SETTINGS
##########################

if COVERAGE:
    extra_compile_args.append('-fprofile-arcs')
    extra_compile_args.append('-ftest-coverage')
    extra_link_args.append('-lgcov')

if UNOPTIMIZED:
    extra_compile_args.append('-O0')

################################################################################
# STATIC SSL LINKING BUILD SETTINGS
################################################################################

if STATIC_SSL:
    extra_objects.extend(
        [SSL_LIB_PATH + 'libssl.a', SSL_LIB_PATH + 'libcrypto.a'])
    libraries.remove('ssl')
    libraries.remove('crypto')
    library_dirs.remove('/usr/local/opt/openssl/lib')

################################################################################
# PLATFORM SPECIFIC BUILD SETTINGS
################################################################################

if WINDOWS:
    AEROSPIKE_C_TARGET = AEROSPIKE_C_HOME
    tree = ET.parse(f"{AEROSPIKE_C_TARGET}/vs/aerospike/packages.config")
    packages = tree.getroot()
    package = packages[0]
    c_client_dependencies_version = package.attrib["version"]

if DARWIN:
    # ---------------------------------------------------------------------------
    # Mac Specific Compiler and Linker Settings
    # ---------------------------------------------------------------------------
    extra_compile_args = extra_compile_args + [
        '-D_DARWIN_UNLIMITED_SELECT'
    ]

    AEROSPIKE_C_TARGET = AEROSPIKE_C_HOME + '/target/Darwin-' + machine

elif LINUX:
    # ---------------------------------------------------------------------------
    # Linux Specific Compiler and Linker Settings
    # ---------------------------------------------------------------------------
    extra_compile_args = extra_compile_args + [
        '-rdynamic', '-finline-functions'
    ]
    libraries = libraries + ['rt']
    AEROSPIKE_C_TARGET = AEROSPIKE_C_HOME + '/target/Linux-' + machine
elif WINDOWS:
    libraries.clear()
    extra_compile_args.append("-DAS_SHARED_IMPORT")
    include_dirs.append(f"{AEROSPIKE_C_TARGET}/vs/packages/aerospike-client-c-dependencies.{c_client_dependencies_version}/build/native/include")
else:
    print("error: OS not supported:", PLATFORM, file=sys.stderr)
    sys.exit(8)

include_dirs = include_dirs + [
    '/usr/local/opt/openssl/include',

]
if not WINDOWS:
    include_dirs.append(AEROSPIKE_C_TARGET + '/include')
    extra_objects = extra_objects + [
        AEROSPIKE_C_TARGET + '/lib/libaerospike.a'
    ]
else:
    include_dirs.append(AEROSPIKE_C_TARGET + '/src/include')
    library_dirs.append(f"{AEROSPIKE_C_TARGET}/vs/packages/aerospike-client-c-dependencies.{c_client_dependencies_version}/build/native/lib/x64/Release")
    # Needed for linking the Python client with the C client
    extra_objects.append(AEROSPIKE_C_TARGET + "/vs/x64/Release/aerospike.lib")

os.putenv('CPATH', ':'.join(include_dirs))
os.environ['CPATH'] = ':'.join(include_dirs)

################################################################################
# SETUP
################################################################################

# Get the long description from the relevant file
with io.open(os.path.join(CWD, 'README.rst'), "r", encoding='utf-8') as f:
    long_description = f.read()

# Get the version from the relevant file
with io.open(os.path.join(CWD, 'VERSION'), "r", encoding='utf-8') as f:
    version = f.read()

BASEPATH = os.path.dirname(os.path.abspath(__file__))
CCLIENT_PATH = os.path.join(BASEPATH, 'aerospike-client-c')


class CClientBuild(build):

    def run(self):
        if self.force == 1:
            # run original c-extension clean task
            # clean.run(self)
            cmd = [
                'make',
                'clean'
            ]
            def clean():
                call(cmd, cwd=CCLIENT_PATH)
            self.execute(clean, [], 'Clean core aerospike-client-c previous builds')

        os.putenv('LD_LIBRARY_PATH', ':'.join(library_dirs))
        os.environ['LD_LIBRARY_PATH'] = ':'.join(library_dirs)
        os.putenv('DYLD_LIBRARY_PATH', ':'.join(library_dirs))
        os.environ['DYLD_LIBRARY_PATH'] = ':'.join(library_dirs)
        # build core client
        if WINDOWS:
            cmd = [
                'msbuild',
                'vs/aerospike.sln',
                '/property:Configuration=Release'
            ]
        else:
            cmd = [
                'make',
                'V=' + str(self.verbose),
            ]
            if UNOPTIMIZED:
                cmd.append('O=0')

        def compile():
            print(cmd, library_dirs, libraries)
            call(cmd, cwd=CCLIENT_PATH)

        self.execute(compile, [], 'Compiling core aerospike-client-c')
        # run original c-extension build code
        build.run(self)

        # For debugging in macOS, we need to generate and include the debug info for the CPython
        # extension in the wheel, since this isn't done automatically
        if DARWIN and INCLUDE_DSYM:
            print("Generating debug information on macOS")
            shared_library_paths = glob.glob(pathname="**/aerospike*.so", recursive=True)

            # Sanity check
            print(f"List of shared libraries: {shared_library_paths}")
            if len(shared_library_paths) > 1:
                print("error: only one shared library should be present.", file=sys.stderr)
                exit(1)

            shared_library_path = shared_library_paths[0]
            run(["dsymutil", shared_library_path], check=True)

            dsym_path = f"{shared_library_path}.dSYM"
            print("Including debug information with wheel")
            extra_objects.append(dsym_path)

class CClientClean(clean):

    def run(self):
        # run original c-extension clean task
        # clean.run(self)
        cmd = [
            'make',
            'clean'
        ]

        def clean():
            call(cmd, cwd=CCLIENT_PATH)

        self.execute(clean, [], 'Clean core aerospike-client-c')


setup(
    version=version.strip(),
    # Data files
    ext_modules=[
        Extension(
            # Extension Name
            'aerospike',

            # Source Files
            [
                'src/main/aerospike.c',
                'src/main/exception.c',
                'src/main/log.c',
                'src/main/client/type.c',
                'src/main/client/apply.c',
                'src/main/client/bit_operate.c',
                'src/main/client/cdt_list_operate.c',
                'src/main/client/cdt_map_operate.c',
                'src/main/client/hll_operate.c',
                'src/main/client/expression_operations.c',
                'src/main/client/cdt_operation_utils.c',
                'src/main/client/close.c',
                'src/main/client/connect.c',
                'src/main/client/exists.c',
                'src/main/client/exists_many.c',
                'src/main/client/get.c',
                'src/main/client/get_many.c',
                'src/main/client/batch_get_ops.c',
                'src/main/client/select_many.c',
                'src/main/client/info_single_node.c',
                'src/main/client/info_random_node.c',
                'src/main/client/info.c',
                'src/main/client/put.c',
                'src/main/client/operate.c',
                'src/main/client/query.c',
                'src/main/client/remove.c',
                'src/main/client/scan.c',
                'src/main/client/select.c',
                'src/main/client/truncate.c',
                'src/main/client/admin.c',
                'src/main/client/udf.c',
                'src/main/client/sec_index.c',
                'src/main/serializer.c',
                'src/main/client/remove_bin.c',
                'src/main/query/type.c',
                'src/main/query/apply.c',
                'src/main/query/add_ops.c',
                'src/main/query/paginate.c',
                'src/main/query/get_parts.c',
                'src/main/query/foreach.c',
                'src/main/query/results.c',
                'src/main/query/select.c',
                'src/main/query/where.c',
                'src/main/query/execute_background.c',
                'src/main/scan/type.c',
                'src/main/scan/foreach.c',
                'src/main/scan/results.c',
                'src/main/scan/select.c',
                'src/main/scan/execute_background.c',
                'src/main/scan/apply.c',
                'src/main/scan/add_ops.c',
                'src/main/scan/paginate.c',
                'src/main/scan/get_parts.c',
                'src/main/geospatial/type.c',
                'src/main/geospatial/wrap.c',
                'src/main/geospatial/unwrap.c',
                'src/main/geospatial/loads.c',
                'src/main/geospatial/dumps.c',
                'src/main/policy.c',
                'src/main/conversions.c',
                'src/main/convert_expressions.c',
                'src/main/policy_config.c',
                'src/main/calc_digest.c',
                'src/main/predicates.c',
                'src/main/tls_config.c',
                'src/main/global_hosts/type.c',
                'src/main/nullobject/type.c',
                'src/main/cdt_types/type.c',
                'src/main/key_ordered_dict/type.c',
                'src/main/client/set_xdr_filter.c',
                'src/main/client/get_expression_base64.c',
                'src/main/client/get_cdtctx_base64.c',
                'src/main/client/get_nodes.c',
                'src/main/convert_partition_filter.c',
                'src/main/client/get_key_partition_id.c',
                'src/main/client/batch_write.c',
                'src/main/client/batch_operate.c',
                'src/main/client/batch_remove.c',
                'src/main/client/batch_apply.c',
                'src/main/client/batch_read.c',
                'src/main/client/metrics.c'
            ],

            # Compile
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,

            # Link
            library_dirs=library_dirs,
            libraries=libraries,
            extra_objects=extra_objects,
            extra_link_args=extra_link_args,
        )
    ],
    package_data={
        "aerospike-stubs": [
            "__init__.pyi",
            "aerospike.pyi",
            "exception.pyi",
            "predicates.pyi",
        ]
    },
    packages=['aerospike_helpers', 'aerospike_helpers.operations', 'aerospike_helpers.batch',
              'aerospike_helpers.expressions',
              'aerospike_helpers.metrics',
              'aerospike-stubs'],
    cmdclass={
        'build': CClientBuild,
        'clean': CClientClean
    }
)



from parver import Version
import versioningit
from typing import Any, Dict, Union
import pathlib

import versioningit.basics
import versioningit.git
import os

# Take in <version> and <string> as input
# If local version identifier doesn't exist, append +<string> to <version>
# Example: 15.0.0rc1 and dsym as input -> 15.0.0rc1+dsym
# If it does exist, append <version> with .<string>
# Then output new version as str
def append_to_local(version_str: str, value: str) -> str:
    version = Version.parse(version_str, strict=True)
    if version.local == None:
        new_local = value
    else:
        new_local = f"{version.local}.{value}"

    version = version.replace(local=new_local)
    return version.__str__()

def my_vcs(
        project_dir: Union[str, pathlib.Path],
        params: Dict[str, Any]
) -> versioningit.VCSDescription:
    vcs_description = versioningit.git.describe_git(
        project_dir=project_dir,
        params=params
    )
    if vcs_description.state == "exact":
        # We don't want the format step to be skipped
        # Workaround: https://github.com/jwodder/versioningit/issues/42#issuecomment-1235573432
        vcs_description.state = "exact_"
    return vcs_description

def my_format(
        description: versioningit.VCSDescription,
        base_version: str,
        next_version: str,
        params: Dict[str, Any]
) -> str:
    # Even if the repository state matches a tag, we always need to label the version if it's unoptimized or includes
    # dsym for macOS
    if description.state != "exact_":
        version_str = versioningit.basics.basic_format(
            description=description,
            base_version=base_version,
            next_version=next_version,
            params=params
        )
    else:
        version_str = base_version

    if os.getenv("UNOPTIMIZED"):
        version_str = append_to_local(version_str, "unoptimized")
    if os.getenv("INCLUDE_DSYM"):
        version_str = append_to_local(version_str, "dsym")

    return version_str
